#!/usr/bin/env python
#coding:utf-8

##################################################
# MUBLAS Analyzer ver.1.5.20
##################################################

debug = 0

##############################################################
# header
##############################################################
import os
import sys
import csv
import string
import operator
import argparse
import numpy
import math
import subprocess
from datetime import datetime
from scipy import stats
import logging
from logging import getLogger,StreamHandler,DEBUG
logger = getLogger(__name__)
handler = logging.FileHandler('mublas_analyzer.log', 'a+')
handler.setLevel(DEBUG)
logger.setLevel(DEBUG)
logger.addHandler(handler)

G_gnuplot_script_header = "#gnuplot script\n" \
	+ 'set term postscript eps enhanced color\n' \
	+ 'set style line 11 lt 1 lw 1.2 ps 1.0 pt 6 lc rgb "#CC3300" # red\n' \
	+ 'set style line 21 lt 1 lw 1.2 ps 1.0 pt 2 lc rgb "#009900" # green\n' \
	+ 'set style line 31 lt 1 lw 1.2 ps 1.0 pt 14 lc rgb "#CC9900" # orange\n' \
	+ 'set style line 32 lt 2 lw 1.2 ps 1.0 pt 15 lc rgb "#CC9900" # orange\n' \
	+ 'set style line 41 lt 1 lw 1.2 ps 1.0 pt 12 lc rgb "#3366CC" # blue\n' \
	+ 'set style line 42 lt 2 lw 1.2 ps 1.0 pt 1 lc rgb "#3366CC" # blue\n' \
	+ 'set style line 81 lt 1 lw 1.2 ps 1.0 pt 13 lc rgb "#FF0099" # pink \n' \
	+ 'set style line 82 lt 2 lw 1.2 ps 1.0 pt 4 lc rgb "#FF0099" # pink \n' \
	+ 'set style line 91 lt 1 lw 1.2 ps 1.0 pt 15 lc rgb "#3399FF" # sky \n' \
	+ 'set style line 92 lt 2 lw 1.2 ps 1.0 pt 6 lc rgb "#3399FF" # sky \n' \
	+ 'set size 0.45,0.65\n' \
	+ 'set grid\n' \
	+ 'set xrange [0:]\n' \
	+ 'set yrange [0:]\n' \
	+ 'set key below\n'

G_dat_header = "%%M\tN\tK\tsec\tGFlops\tGB/s\titer\ttx\tty\tth\tyx\tsb\tbd\tbm\tWrpOcp\tBlkOcp\tGrdOcp\tcand\tkernel#"

##############################################################
# functions:
##############################################################
def ceiling (n, c):
	return math.ceil (n/c) * c

def import_csv_file (file_in):
	with open (file_in, 'r') as dat:
		csv_dat = list(csv.reader(dat, delimiter='\t'))
	for row in range(len(csv_dat)):
		for cell in range(len(csv_dat[row])):
			try:
				csv_dat[row][cell] = float(csv_dat[row][cell])
			except:
				pass
	return csv_dat

def export_csv_file (csv_dat, file_ou):
	for row in csv_dat:
		row = [str(text) for text in row]
		cmd = 'echo "' + '\t'.join(row) + '" >> ' + file_ou
		do (cmd)

def sort_csv_dat(csv_dat, key):
	header = csv_dat[0]
	body = sorted (csv_dat[1:], key=operator.itemgetter(header.index(key)), reverse=True)
	body.insert (0, header)
	return body

def list_to_str (list_dat):
	dat = ""
	for x in list_dat:
		dat += str(x) + "\t"
	return dat

def do (command):
	if (debug == 1):
		print "[cmd] " + command
	elif (debug == 2):
		print "[cmd] " + command
		os.system (command)
	else:
		os.system (command)

def debug_msg (ah, msg):
	logger.debug (datetime.now().strftime("%Y%m%d%H%M") + " < mublas_analyzer > " + msg)
	#print ("< mublas_analyzer > " + msg)

def genplotscript (ah, param, x, y, color):
	plottitle = ah['func_dname'].upper() + ' on ' + ah['device']
	plottitle = plottitle.replace('_', '-')
	plottitle = plottitle.replace('TRMV-L', 'TRMV-LNN')
	text = "\n\n#" + param + "\n" \
		+ 'set title "' + plottitle + '"\n' \
		+ 'set nokey \n' \
		+ 'set ylabel "Performance [GFlops]" \n' \
		+ 'set xlabel "' + param + '"\n'
	dim_n = ah['dim_start']
	text += 'set output "fig_' + ah['file_base'] + '_' + param + '.eps"\n' 
	text += 'plot \\\n' 
	while (dim_n < ah['dim_stop']):
		# ---------------
		filename = ah['file_base'] + "_n" + str(dim_n) + ".dat"
		text += '"' + filename + '" using ' + str(x) + ':' + str(y) + ' every 1 w points ls ' + str(color) + ',\\\n'
		# ---------------
		dim_n += ah['dim_step']
	text += '"' + filename + '" using ' + str(x) + ':' + str(y) + ' every 1 w points ls ' + str(color) + '\n'
	return text

def genplotscript_n (ah, param, x, y, color):
	plottitle = ah['func_dname'].upper() + ' on ' + ah['device']
	plottitle = plottitle.replace('_', '-')
	plottitle = plottitle.replace('TRMV-L', 'TRMV-LNN')
	text = "\n\n#" + param + "\n" \
		+ 'set title "' + plottitle + '"\n' \
		+ 'set nokey \n' \
		+ 'set ylabel "Performance [GFlops]" \n' \
		+ 'set xlabel "' + param + '"\n'
	dim_n = ah['dim_start']
	while (dim_n <= ah['dim_stop']):
		text += 'set output "fig_' + ah['file_base'] + '_n' + str(dim_n) + '_' + param + '.eps"\n' 
		text += 'plot \\\n' 
		# ---------------
		filename = ah['file_base'] + "_n" + str(dim_n) + ".dat"
		text += '"' + filename + '" using ' + str(x) + ':' + str(y) + ' every 1 w points ls ' + str(color) + '\n'
		# ---------------
		dim_n += ah['dim_step']
	return text

def genplotscript_n_o (ah):
	plottitle = ah['func_dname'].upper() + ' on ' + ah['device']
	plottitle = plottitle.replace('_', '-')
	plottitle = plottitle.replace('TRMV-L', 'TRMV-LNN')
	text = "\n\n#\n" \
		+ 'set title "' + plottitle + '"\n' \
		+ 'set key below \n' \
		+ 'set ylabel "Performance [GFlops]" \n' \
		+ 'set xlabel "Occupancy [%]"\n'
	dim_n = ah['dim_start']
	while (dim_n <= ah['dim_stop']):
		text += 'set output "fig_' + ah['file_base'] + '_n' + str(dim_n) + '_occupancies.eps"\n' 
		text += 'plot \\\n' 
		# ---------------
		text += '"' + ah['file_base'] + '_n' + str(dim_n) + '.dat" using ($14*100):5 title "wo" w points ls 11,\\\n'
		text += '"' + ah['file_base'] + '_n' + str(dim_n) + '.dat" using ($15*100):5 title "bo" w points ls 31,\\\n'
		text += '"' + ah['file_base'] + '_n' + str(dim_n) + '.dat" using ($16*100):5 title "go" w points ls 21\n'
		# ---------------
		dim_n += ah['dim_step']
	return text

def gen_analyzer_info (ah, filename):
	file = open (filename, 'w')
	print >> file, ah
	file.close()

def total_num_of_execution (ah):
	file_in = ah['file_dir'] + ah['file_base'] + "_mode_t.dat"
	f = open (file_in)
	lines = f.readlines ()
	f.close ()
	count = 0
	for line in lines:
		cols = line.split(",")
		for col in cols[1:]:
			if col=="\n":
				break
			count += 1
	ah['totalnumber'] = count

def testing_exec (blas, ah, file_ou):
	cmd = 'echo "' + G_dat_header + '" > ' + file_ou
	do (cmd)
	dim = ah['dim_start']
	while (dim <= ah['dim_stop']):
		cmd = "./testing_" + blas + "_" + ah['routine'] + " --nodisp=1 --mode=p" \
			+ " --devno=" + str(ah['devno']) + " --uplo=" + ah['uplo'] + " --trans=" + ah['trans'] + " --diag=" + ah['diag'] \
			+ " --kernel=" + str(ah['kernel']) \
			+ " --range=" + str(dim) + ":" + str(dim) + ":1" \
			+ " >> " + file_ou
		do (cmd)
		dim += ah['dim_step']

def testing_exec_txty (blas, ah, file_ou, tx, ty):
	cmd = 'echo "' + G_dat_header + '" > ' + file_ou
	do (cmd)
	dim = ah['dim_start']
	while (dim <= ah['dim_stop']):
		cmd = "./testing_" + blas + "_" + ah['routine'] + " --nodisp=1 --mode=p" \
			+ " --devno=" + str(ah['devno']) + " --uplo=" + ah['uplo'] + " --trans=" + ah['trans'] + " --diag=" + ah['diag'] \
			+ " --kernel=" + str(ah['kernel']) \
			+ " --range=" + str(dim) + ":" + str(dim) + ":1" \
			+ " --tx=" + str(tx) + " --ty=" + str(ty) \
			+ " >> " + file_ou
		do (cmd)
		dim += ah['dim_step']

##############################################################
# Evaluation
##############################################################
def do_evaluation (ah):
	debug_msg (ah, "do_evaluation: start.")

	file_ou = ah['file_dir'] + ah['file_base'] + "_eval_cublas.dat"
	debug_msg (ah, 'creating "' + file_ou + '"')
	testing_exec ("cublas", ah, file_ou)

	# open Tx&Ty file
	file_io = ah['file_dir'] + ah['file_base'] + "_eval_mublas_auto_consecutive.dat"
	debug_msg (ah, 'creating "' + file_io + '"')
	testing_exec ("mublas", ah, file_io)
	csv_dat = import_csv_file (file_io)
	dat_dim_n = [row[1] for row in csv_dat]
	dat_tx = [row[7] for row in csv_dat]
	dat_ty = [row[8] for row in csv_dat]
	file_ou1 = ah['file_dir'] + ah['file_base'] + "_eval_mublas_auto_static.dat"
	file_ou2 = ah['file_dir'] + ah['file_base'] + "_eval_mublas_auto.dat"
	debug_msg (ah, 'creating "' + file_ou1 + '" and "' + file_ou2 + '"')
	cmd = "echo '%%' > " + file_ou1
	do (cmd)
	cmd = "echo '%%' > " + file_ou2
	do (cmd)
	for i in range(len(dat_dim_n)):
		cmd = "./testing_mublas_" + ah['routine'] + " --nodisp=1 --mode=p" \
		+ " --devno=" + str(ah['devno']) + " --uplo=" + ah['uplo'] + " --trans=" + ah['trans'] + " --diag=" + ah['diag'] \
		+ " --kernel=" + str(ah['kernel']) \
		+ " --tx=" + str(dat_tx[i]) + " --ty=" + str(dat_ty[i]) \
		+ " --range=" + str(dat_dim_n[i]) + ":" + str(dat_dim_n[i]) + ":1 >> " \
		+ file_ou1
		do (cmd)
		cmd = "./testing_mublas_" + ah['routine'] + " --nodisp=1 --mode=p" \
		+ " --devno=" + str(ah['devno']) + " --uplo=" + ah['uplo'] + " --trans=" + ah['trans'] + " --diag=" + ah['diag'] \
		+ " --kernel=" + str(ah['kernel']) \
		+ " --range=" + str(dat_dim_n[i]) + ":" + str(dat_dim_n[i]) + ":1 >> " \
		+ file_ou2
		do (cmd)
	debug_msg (ah, "do_evaluation: done.")

def do_evaluation_txty (ah):
	debug_msg (ah, "do_evaluation: start.")

	file_io = ah['file_dir'] + ah['file_base'] + "_eval_mublas_tx32_ty32.dat"
	debug_msg (ah, 'creating "' + file_io + '"')
	testing_exec_txty ("mublas", ah, file_io, 32, 32)

	file_io = ah['file_dir'] + ah['file_base'] + "_eval_mublas_tx32_ty4.dat"
	debug_msg (ah, 'creating "' + file_io + '"')
	testing_exec_txty ("mublas", ah, file_io, 32, 4)

	file_io = ah['file_dir'] + ah['file_base'] + "_eval_mublas_tx32_ty8.dat"
	debug_msg (ah, 'creating "' + file_io + '"')
	testing_exec_txty ("mublas", ah, file_io, 32, 8)


#	file_io = ah['file_dir'] + ah['file_base'] + "_eval_mublas_tx128_ty1.dat"
#	debug_msg (ah, 'creating "' + file_io + '"')
#	testing_exec_txty ("mublas", ah, file_io, 128, 1)

#	file_io = ah['file_dir'] + ah['file_base'] + "_eval_mublas_tx128_ty2.dat"
#	debug_msg (ah, 'creating "' + file_io + '"')
#	testing_exec_txty ("mublas", ah, file_io, 128, 2)

#	file_io = ah['file_dir'] + ah['file_base'] + "_eval_mublas_tx256_ty1.dat"
#	debug_msg (ah, 'creating "' + file_io + '"')
#	testing_exec_txty ("mublas", ah, file_io, 256, 1)

	file_io = ah['file_dir'] + ah['file_base'] + "_eval_mublas_tx32_ty16.dat"
	debug_msg (ah, 'creating "' + file_io + '"')
	testing_exec_txty ("mublas", ah, file_io, 32, 16)

#	file_io = ah['file_dir'] + ah['file_base'] + "_eval_mublas_tx64_ty16.dat"
#	debug_msg (ah, 'creating "' + file_io + '"')
#	testing_exec_txty ("mublas", ah, file_io, 64, 16)

#	file_io = ah['file_dir'] + ah['file_base'] + "_eval_mublas_tx16_ty16.dat"
#	debug_msg (ah, 'creating "' + file_io + '"')
#	testing_exec_txty ("mublas", ah, file_io, 16, 16)


##############################################################
# Execution
##############################################################
#DIMRANGE
def do_evaluation_for_all_tbsizes (ah):
	debug_msg (ah, "do_evaluation_for_all_tbsizes: start.")
	cmd = "mkdir -p " + ah['file_dir']
	do (cmd)

	file_io = ah['file_dir'] + ah['file_base'] + "_mode_t.dat"
	debug_msg (ah, 'creating "' + file_io + '"')
	cmd = "./testing_mublas_" + ah['routine'] + " --nodisp=1 --mode=t" \
		+ " --devno=" + str(ah['devno']) + " --uplo=" + ah['uplo'] + " --trans=" + ah['trans'] + " --diag=" + ah['diag'] \
		+ " --kernel=" + str(ah['kernel']) \
		+ " --range=" + str(ah['dim_start']) + ":" + str(ah['dim_stop']) + ":" + str(ah['dim_step']) \
		+ " > " + file_io
	do (cmd)

	file_ou = ah['file_dir'] + ah['file_base'] + "_info.dat"
	debug_msg (ah, 'creating "' + file_ou + '"')
	gen_analyzer_info (ah, file_ou)
	total_num_of_execution (ah)

	f = open (file_io)
	lines = f.readlines ()
	f.close ()
	for line in lines:
		cols = line.split(",")
		dim = cols[0]
		file_ou = ah['file_dir'] + ah['file_base'] + "_n" + str(dim) + ".dat"
		cmd = 'echo "' + G_dat_header + '" > ' + file_ou
		do (cmd)
	
	count = 0
	for line in lines:
		cols = line.split(",")
		dim = cols[0]
		file_ou = ah['file_dir'] + ah['file_base'] + "_n" + str(dim) + ".dat"
		for col in cols[1:]:
			if col=="\n":
				break
			count += 1
			th = col.split("x")
			tx = th[0]
			ty = th[1]
			debug_msg (ah, 'creating "' + file_ou + '" (' + str(count) + '/' + str(ah['totalnumber']) + ')')
			cmd = "./testing_mublas_" + ah['routine'] + " --nodisp=1 --mode=p" \
				+ " --devno=" + str(ah['devno']) + " --uplo=" + ah['uplo'] + " --trans=" + ah['trans'] + " --diag=" + ah['diag'] \
				+ " --kernel=" + str(ah['kernel']) \
				+ " --range=" + str(dim) + ":" + str(dim) + ":1" \
				+ " --tx=" + str(tx) + " --ty=" + str(ty) \
				+ " >> " + file_ou
			do (cmd)
	debug_msg (ah, "do_evaluation_for_all_tbsizes: done.")

#DIMRANGE
def gen__statistics (ah):
	debug_msg (ah, "gen__statistics: start.")
	file_ou = ah['file_dir'] + ah['file_base'] + "_eval_statistics.dat"
	debug_msg (ah, 'creating "' + file_ou + '"')
	cmd = 'echo "%%N\tQ1\tmin\tmax\tQ3\tmean\tmedian\tcand" > ' + file_ou
	do (cmd)
	dim_n = ah['dim_start']
	while (dim_n <= ah['dim_stop']):
		file_in = ah['file_dir'] + ah['file_base'] + "_n" + str(dim_n) + ".dat"
		csv_dat = import_csv_file (file_in)
		dat_perf = [row[4] for row in csv_dat]
		dat_perf.pop(0)
		cmd = 'echo "' \
			+ str(dim_n) + '\t' \
			+ str(stats.scoreatpercentile(dat_perf, 25)) + '\t' \
			+ str(numpy.min(dat_perf)) + '\t' \
			+ str(numpy.max(dat_perf)) + '\t' \
			+ str(stats.scoreatpercentile(dat_perf, 75)) + '\t' \
			+ str(numpy.mean(dat_perf)) + '\t' \
			+ str(numpy.median(dat_perf)) + '\t' \
			+ str(ah['totalnumber']) \
			+ '" >> ' + file_ou
		do (cmd)
		dim_n += ah['dim_step']
	debug_msg (ah, "gen__statistics: done.")
	
#DIMRANGE
def gen__mublas_offline (ah):
	debug_msg (ah, "gen__mublas_offline: start.")
	file_ou = ah['file_dir'] + ah['file_base'] + "_eval_mublas_best.dat"
	debug_msg (ah, 'creating "' + file_ou + '"')
	cmd = 'echo "' + G_dat_header + '" > ' + file_ou
	do (cmd)
	dim_n = ah['dim_start']
	while (dim_n <= ah['dim_stop']):
		file_in = ah['file_dir'] + ah['file_base'] + "_n" + str(dim_n) + ".dat"
		csv_dat = import_csv_file (file_in)
		csv_sorted = sort_csv_dat (csv_dat, 'GFlops')
		cmd = 'echo "' + list_to_str (csv_sorted[1]) + '" >> ' + file_ou
		do (cmd)
		dim_n += ah['dim_step']
	debug_msg (ah, "gen__mublas_offline: done.")

def do_recompile (ah):
	debug_msg (ah, "do_recompile: start.")
	cmd = "cd ../;make clean --silent;make -j9 --silent;cd analyzer"
	do (cmd)
	debug_msg (ah, "do_recompile: done.")

##############################################################
# Plot
##############################################################
def do_plot (ah):
	debug_msg (ah, "do_plot: start.")

	# script file (header)
	script = G_gnuplot_script_header  
	script += genplotscript_n_o (ah)
	script += genplotscript_n (ah, 'Warp-Occupancy', 14, 5, 11)
	script += genplotscript_n (ah, 'Block-Occupancy', 15, 5, 31)
	script += genplotscript_n (ah, 'Grid-Occupancy', 16, 5, 21)
	script += genplotscript (ah, 'tx', 8, 5, 41)
	script += genplotscript (ah, 'ty', 9, 5, 41)
	script += genplotscript (ah, 'th', 10, 5, 41)
	script += genplotscript (ah, 'tyx', 11, 5, 41)

	# online vs offline
	plottitle = ah['func_dname'].upper() + ' on ' + ah['device']
	plottitle = plottitle.replace('_', '-')
	plottitle = plottitle.replace('TRMV-L', 'TRMV-LNN')
#	script = G_gnuplot_script_header
	script += "\n\n#stat\n" \
		+ 'set title "' + plottitle + '"\n' \
		+ 'set key below \n' \
		+ 'set xlabel "Problem Size (n)"\n' \
		+ 'set ylabel "Performance [GFlops]" offset 1.5,0\n' \
		+ 'set xrange [0:' + str(ah['dim_stop']) + ']\n' \
		+ 'set yrange [0:*]\n' \
		+ 'set xtics ' +  str(ah['dim_stop']/4) + '\n' \
		+ 'set output "fig_' + ah['file_base'] + '_perf.eps"\n' \
		+ 'set boxwidth 0.4 relative\n' \
		+ 'set style fill empty\n' \
		+ 'plot \\\n' \
		+ '"' + ah['file_base'] + '_eval_statistics.dat" using 1:7:7:7:7 notitle w candlesticks ls 21,\\\n' \
		+ '"' + ah['file_base'] + '_eval_statistics.dat" using 1:2:3:4:5 title "all" w candlesticks whiskerbars ls 21,\\\n' \
		+ '"' + ah['file_base'] + '_eval_mublas_auto_static.dat" using 2:5 title "auto-kernel" w linespoints ls 31,\\\n' \
		+ '"' + ah['file_base'] + '_eval_mublas_auto.dat" using 2:5 title "auto" w linespoints ls 11,\\\n' \
		+ '"' + ah['file_base'] + '_eval_cublas.dat" using 2:5 title "cuBLAS" w linespoints ls 42,\\\n' \
		+ '"' + ah['file_base'] + '_eval_mublas_tx32_ty4.dat" using 2:5 title "(32,4)" w linespoints ls 31,\\\n' \
		+ '"' + ah['file_base'] + '_eval_mublas_tx32_ty8.dat" using 2:5 title "(32,8)" w linespoints ls 32,\\\n' \
		+ '"' + ah['file_base'] + '_eval_mublas_tx32_ty16.dat" using 2:5 title "(32,16)" w linespoints ls 33,\\\n' \
		+ '"' + ah['file_base'] + '_eval_mublas_tx32_ty32.dat" using 2:5 title "(32,32)" w linespoints ls 34\n' \
		+ '\n#txty\n' \
		+ 'set output "fig_' + ah['file_base'] + '_txty.eps"\n' \
		+ 'set ylabel "# of threads"\n' \
		+ 'set autoscale y\n' \
		+ 'plot \\\n' \
		+ '"' + ah['file_base'] + '_eval_mublas_auto.dat" using 2:8 title "tx(auto)" w linespoints ls 81,\\\n' \
		+ '"' + ah['file_base'] + '_eval_mublas_auto.dat" using 2:9 title "ty(auto)" w linespoints ls 91,\\\n' \
		+ '"' + ah['file_base'] + '_eval_mublas_best.dat" using 2:8 title "tx(best)" w linespoints ls 82,\\\n' \
		+ '"' + ah['file_base'] + '_eval_mublas_best.dat" using 2:9 title "ty(best)" w linespoints ls 92\n' 


	file_ou = ah['file_dir'] + ah['file_base'] + ".plt"
	debug_msg (ah, 'creating "' + file_ou + '"')
	cmd = "echo '" + script + "' > " + file_ou
	do (cmd)
	os.chdir (ah['file_dir'])
	debug_msg (ah, 'gnuplot mublas_analyzer_' + ah['func_dname'] + '.plt')
	cmd = "gnuplot " + ah['file_base'] + ".plt"
	do (cmd)
	os.chdir ("../../")
	debug_msg (ah, "do_plot: done.")

##############################################################
# main
##############################################################
def main ():
	argvs = sys.argv
	argc = len (argvs)
	parser = argparse.ArgumentParser (description="mublas analyzer")
	parser.add_argument ('--devno', dest='devno', action='store', default='0', help='devno') 
	parser.add_argument ('--routine', dest='routine', action='store', default='', help='routine') 
	parser.add_argument ('--uplo', dest='uplo', action='store', default='', help='uplo') 
	parser.add_argument ('--trans', dest='trans', action='store', default='', help='trans') 
	parser.add_argument ('--diag', dest='diag', action='store', default='', help='diag') 
	parser.add_argument ('--dim', dest='dim', action='store', default='0', help='dim') 
	parser.add_argument ('--dimrange', dest='dimrange', action='store', default='0:0:0', help='dimrange') 
	parser.add_argument ('--process', dest='process', action='store', default='', help='process') 
	parser.add_argument ('--kernel', dest='kernel', action='store', default='0', help='kernel') 
	args = parser.parse_args()

	option = ''
	routine = ''
	if (args.uplo != ''):
		option += args.uplo
	if (args.trans != ''):
		option += args.trans
	if (args.diag != ''):
		option += args.diag
	if (option != ''):
		routine += args.routine + "_" + option.lower()
	else:
		routine += args.routine 

	# get devinfo and create analyzer-handler
	ah = { 
		'devno': int(args.devno), \
		'routine': args.routine.lower(), \
		'uplo': args.uplo.lower(), \
		'trans': args.trans.lower(), \
		'diag': args.diag.lower(), \
		'dim': int(args.dim), \
		'dim_start': int(args.dimrange.split(':')[0]), \
		'dim_stop': int(args.dimrange.split(':')[1]), \
		'dim_step': int(args.dimrange.split(':')[2]), \
		'kernel': int(args.kernel), \
		'cc': '', \
		'device': '', \
		'mublasversion': '', \
		'date': datetime.now().strftime("%Y%m%d%H%M"), \
		'totalnumber' : 0, \
		'file_base': "mublas_analyzer_" + routine.lower(), \
		'file_dir': '', \
		'func_dname': routine.lower() \
	}

	debug_msg (ah, "---------------------------------------------" )

# ----------------------------
#	if (args.process == 'all'):
#		do_recompile (ah)
# ----------------------------

	cmd = "cp ../testing/testing_mublas_devinfo ."
	do (cmd)
	cmd = "cp ../testing/testing_mublas_" + ah['routine'] + " ."
	do (cmd)
	cmd = "cp ../testing/testing_cublas_" + ah['routine'] + " ."
	do (cmd)

	debug_msg (ah, "get device info...")
	cmd = "./testing_mublas_devinfo --devno=" + str(ah['devno']) + " > devinfo"
	do (cmd)
	f = open ("devinfo")
	lines = f.readlines ()
	f.close ()
	ah['cc'] = lines[0].rstrip()
	ah['device'] = lines[1].rstrip().replace(' ', '_')
	ah['mublasversion'] = lines[2].rstrip()
	ah['file_dir'] =  ah['device'] + "/" + routine.lower() + "_" + str(ah['kernel']) + "/"

	if (ah['dim_stop'] == 0 and ah['dim'] == 0):
		debug_msg (ah, "dim error")
		sys.exit()
	if (ah['dim_stop'] == 0):
		debug_msg (ah, "ah[dim_stop] is empty, set ah[dim] to ah[dim_start] and ah[dim_stop]")
		ah['dim_start'] = ah['dim']
		ah['dim_stop'] = ah['dim']
		ah['dim_step'] = 1
	if (ah['dim'] == 0):
		debug_msg (ah, "ah[dim] is empty, set ah[dim_stop] to ah[dim]")
		ah['dim'] = ah['dim_stop']

	if (args.process == 'all'):
		do_evaluation_for_all_tbsizes (ah)
		gen__statistics (ah)
		gen__mublas_offline (ah)
		do_evaluation (ah)
		do_plot (ah)
	elif (args.process == 'reeval'):
		do_evaluation_txty (ah)
		do_evaluation (ah)
		do_plot (ah)
	elif (args.process == 'reevaltxty'):
		do_evaluation_txty (ah)
	elif (args.process == 'replot'):
		do_plot (ah)
# ----------------------------

	cmd = "rm testing_mublas_devinfo devinfo"
	do (cmd)
	cmd = "rm testing_mublas_" + ah['routine'] 
	do (cmd)
	cmd = "rm testing_cublas_" + ah['routine'] 
	do (cmd)

	debug_msg (ah, "---------------------------------------------" )

##############################################################
# main
##############################################################
if __name__ == "__main__":
	main ()

