/* ***********************************************

AXPY

memo:
Actually v2&v4 kernel_numbers are not needed.
v1 kernel_number with AT is enought to achieve best throughput.

  *********************************************** */   

#include "mublas_devel.h"
#include "mublas_axpy_kernel.cu"

/* ***********************************************
	Host Interface
   *********************************************** */   
__host__ int32_t
MUBLAS_AXPY (
	mublasHandle_t* mh,	
	const size_t n,
	const TYPE0 *alpha,
	const TYPE0 *x,
	const size_t incx,
	TYPE0 *y,
	const size_t incy
) {
	uint32_t kernel_number = 999;
	void (*kernel) (const SIZE_T, const TYPE0, const TYPE0*, TYPE0*);
	mublasRecipeDefault (mh);

	// ---------------------------------------------
	// 1. Shortcut
	if (n == 0 || *alpha == ZERO0) 
		return 0;
	// ---------------------------------------------

	// ---------------------------------------------
	// 2. Kernel Configuration 
	mh->n = n;
	mh->N_one = sizeof (TYPE0);
	mh->N_T = 1;
	if      (VEC4) mh->N_T = 4;
	else if (VEC2) mh->N_T = 2;
	if ((((long)x%(128/sizeof(TYPE0)))%16 != 0) || 
		(((long)y%(128/sizeof(TYPE0)))%16 != 0)) mh->N_T = 1;
	if (incx == 1 && incy == 1) {
		kernel_number = 100 + mh->N_T * 10;
		mh->BlkDim = 1;
		mh->func_getShmPerBlk = &getShmPerBlk_AXPY;
		mh->func_getBlkPerGrd = &getBlkPerGrd_AXPY;
		mh->func_cache = cudaFuncCachePreferL1;
		#include "mublas_axpy.incl"
	}
	switch (kernel_number) {
		case 140: kernel = &MUBLAS_AXPY_DEVICE_v4; break;
		case 120: kernel = &MUBLAS_AXPY_DEVICE_v2; break;
		case 110: kernel = &MUBLAS_AXPY_DEVICE_s1; break;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 3. Unsupported operations (CUBLAS)
	if (kernel_number == 999) {
		#if defined (CUBLAS_AXPY)
		cublasHandle_t ch;
		cublasCreate (&ch);
		cublasSetStream (ch, mh->streamId);
		cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
		CUBLAS_AXPY (ch, n, alpha, x, incx, y, incy);
		cublasDestroy (ch);
		#else
		printf ("mublas error: not implemented");
		#endif
		mh->kernelnumber = kernel_number;
		return 0;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 4. Thread-Block Size Adjust and Kernel Launch
	cudaFuncGetAttributes (&mh->funcatb, kernel);
	cudaFuncSetCacheConfig (kernel, mh->func_cache);
	mublas_optimizer (mh);
	dim3 threads = dim3 (mh->Tx, mh->Ty);
	dim3 grid = dim3 (mh->Bx, mh->By);
	if (mh->autotuning!=1) kernel <<< grid, threads, 0, mh->streamId >>> (n, *alpha, x, y);
	// ---------------------------------------------

	mh->kernelnumber = kernel_number;
	return 0;
}

