/* ***********************************************
 
	GEMM for CUDA 

   *********************************************** */   

#define DBWORK

#include "mublas_devel.h"
#include "mublas_gemm_nn_kernel.cu"
#include "mublas_gemm_tn_kernel.cu"
#include "mublas_gemm_nt_kernel.cu"
#include "mublas_gemm_tt_kernel.cu"

/* ***********************************************
	Host Interface
   *********************************************** */   
 
#if defined (DBWORK)
__global__ void mublasDtoDBmatconv (
	const SIZE_T m,
	const SIZE_T n,
	const TYPE0 * __restrict__ Ag,
	const SIZE_T lda,
	TYPE1 *Bg,
	const SIZE_T ldb
) {
	SIZE_T ix, iy;
	const SIZE_T iBx = blockIdx.x;
	const SIZE_T iBy = blockIdx.y;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T nTx = blockDim.x;
	const SIZE_T nTy = blockDim.y;
	ix = iBx * nTx + iTx;
	iy = iBy * nTy + iTy;
	if (ix < m && iy < n) {
#if defined (PREC_DB)
		Bg[iy * ldb + ix] = double_to_cudfreal (Ag[iy * lda + ix]);
#else
		Bg[iy * ldb + ix] = Ag[iy * lda + ix];
#endif
	}
}
#endif

__host__ int32_t
#ifdef MUBLAS_CC
MUBLAS_GEMM_CC (
#else
MUBLAS_GEMM (
	mublasHandle_t* mh,	
#endif
	const char transA, const char transB,
	const size_t m, const size_t n, const size_t k,
	const TYPE0 *alpha,
	const TYPE0 *a, const size_t lda,
	const TYPE0 *b, const size_t ldb,
	const TYPE0 *beta,
	TYPE0 *c, const size_t ldc
) {
#if defined (DBWORK)
	void (*kernel) (const SIZE_T, const SIZE_T, const SIZE_T, const TYPE0, const TYPE1*, const SIZE_T, const TYPE1*, const SIZE_T, const TYPE0, TYPE0*, const SIZE_T);
#else
	void (*kernel) (const SIZE_T, const SIZE_T, const SIZE_T, const TYPE0, const TYPE0*, const SIZE_T, const TYPE0*, const SIZE_T, const TYPE0, TYPE0*, const SIZE_T);
#endif
	size_t kernel_number = 999;
	size_t use_vector = 0;
	dim3 threads, grid;

#ifdef MUBLAS_CC
	mublasHandle_t *mh;
	mh = (mublasHandle_t *)malloc(sizeof(mublasHandle_t));
	mublasCreateSimple (mh);
#else
	mublasRecipeDefault (mh);
#endif
	
	// ---------------------------------------------
	// 1.
	// ---------------------------------------------
#if defined (DBWORK)
	size_t Bx, By;
	size_t row, col;
	threads = dim3 (16, 16);
	// mat A
	TYPE1 *dev_Ab;
	row = (transA == 'N' || transA == 'n') ? m : k;
	col = (transA == 'N' || transA == 'n') ? k : m;
	size_t lda_ = ceilingEx (row, 2);
	cudaMalloc ((void **) &dev_Ab, sizeof (TYPE1)* lda_ * col);
	Bx = ceilingEx (row / 16., 1);
	By = ceilingEx (col / 16., 1);
	grid = dim3 (Bx, By);
	mublasDtoDBmatconv <<< grid, threads >>> (row, col, a, lda, dev_Ab, lda_);

	// mat B
	TYPE1 *dev_Bb;
	row = (transB == 'N' || transB == 'n') ? k : n;
	col = (transB == 'N' || transB == 'n') ? n : k;
	size_t ldb_ = ceilingEx (row, 2);
	cudaMalloc ((void **) &dev_Bb, sizeof (TYPE1)* ldb_ * col);
	Bx = ceilingEx (row / 16., 1);
	By = ceilingEx (col / 16., 1);
	grid = dim3 (Bx, By);
	mublasDtoDBmatconv <<< grid, threads >>> (row, col, b, ldb, dev_Bb, ldb_);
#endif

	// ---------------------------------------------
	// 2. Kernel Configuration 
	mh->m = m;
	mh->n = n;
	mh->k = k;
	mh->N_one = sizeof (TYPE0);
	mh->N_T = 1;
	if (VEC4) {
		if (
			(((long)lda%(128/sizeof(TYPE0)))%16 == 0) && (m%64 == 0) &&
			(((long)ldb%(128/sizeof(TYPE0)))%16 == 0) && (n%64 == 0) &&
			(((long)ldc%(128/sizeof(TYPE0)))%16 == 0) && (k%64 == 0)
		) {
			mh->N_T = 4;
		} else {
			mh->N_T = 2;
		}
	} else if (VEC2) {
		mh->N_T = 2;
	}
	use_vector = 1;
	// ---------------------------------------------

	// ---------------------------------------------
	// 3. Kernel Select
	if (transB == 'N' || transB == 'n') {
		if (transA == 'N' || transA == 'n') {
			// NN-kernel
			mh->BlkDim = 2;
			mh->func_cache = cudaFuncCachePreferShared;
			if (mh->N_T == 4 && use_vector) {
				kernel_number = 140;
				mh->TxStatic = 8;
				mh->TyStatic = 8;
				mh->func_getShmPerBlk = &getShmPerBlk_GEMM_NN_rb8x8;
				mh->func_getBlkPerGrd = &getBlkPerGrd_GEMM_NN_rb8x8;
				#ifndef PREC_DB
				kernel = &MUBLAS_GEMM_NN_DEVICE_tb8x8_rb8x8_v4_maxwell;
				#endif
			}
			else if (mh->N_T == 2 && use_vector) {
				kernel_number = 120;
				mh->TxStatic = 16;
				mh->TyStatic = 16;
				mh->func_getShmPerBlk = &getShmPerBlk_GEMM_NN_rb2x2;
				mh->func_getBlkPerGrd = &getBlkPerGrd_GEMM_NN_rb2x2;
				kernel = &MUBLAS_GEMM_NN_DEVICE_tb16x16_rb2x2_v2_maxwell;
			} else {
				kernel_number = 110;
				mh->TxStatic = 16;
				mh->TyStatic = 16;
				mh->func_getShmPerBlk = &getShmPerBlk_GEMM_NN_rb1x1;
				mh->func_getBlkPerGrd = &getBlkPerGrd_GEMM_NN_rb1x1;
				kernel = &MUBLAS_GEMM_NN_DEVICE_tb16x16_rb1x1_s1_maxwell;
			}
		} else {
			// TN-kernel
			if (mh->N_T == 2 && use_vector) {
				kernel_number = 220;
				mh->TxStatic = 16;
				mh->TyStatic = 16;
				mh->func_getShmPerBlk = &getShmPerBlk_GEMM_TN_rb2x2;
				mh->func_getBlkPerGrd = &getBlkPerGrd_GEMM_TN_rb2x2;
				kernel = &MUBLAS_GEMM_TN_DEVICE_tb16x16_rb2x2_v2_maxwell;
			} else {
				kernel_number = 210;
				mh->TxStatic = 16;
				mh->TyStatic = 16;
				mh->func_getShmPerBlk = &getShmPerBlk_GEMM_TN_rb1x1;
				mh->func_getBlkPerGrd = &getBlkPerGrd_GEMM_TN_rb1x1;
				kernel = &MUBLAS_GEMM_TN_DEVICE_tb16x16_rb1x1_s1_maxwell;
			}
		}
	} else {
		if (transA == 'N' || transA == 'n') {
			// NT-kernel
			if (mh->N_T == 2 && use_vector) {
				kernel_number = 320;
				mh->TxStatic = 16;
				mh->TyStatic = 16;
				mh->func_getShmPerBlk = &getShmPerBlk_GEMM_NT_rb2x2;
				mh->func_getBlkPerGrd = &getBlkPerGrd_GEMM_NT_rb2x2;
				kernel = &MUBLAS_GEMM_NT_DEVICE_tb16x16_rb2x2_v2_maxwell;
			} else {
				kernel_number = 310;
				mh->TxStatic = 16;
				mh->TyStatic = 16;
				mh->func_getShmPerBlk = &getShmPerBlk_GEMM_NT_rb1x1;
				mh->func_getBlkPerGrd = &getBlkPerGrd_GEMM_NT_rb1x1;
				kernel = &MUBLAS_GEMM_NT_DEVICE_tb16x16_rb1x1_s1_maxwell;
			}
		} else {
			// TT-kernel
			if (mh->N_T == 2 && use_vector) {
				kernel_number = 420;
				mh->TxStatic = 16;
				mh->TyStatic = 16;
				mh->func_getShmPerBlk = &getShmPerBlk_GEMM_TT_rb2x2;
				mh->func_getBlkPerGrd = &getBlkPerGrd_GEMM_TT_rb2x2;
				kernel = &MUBLAS_GEMM_TT_DEVICE_tb16x16_rb2x2_v2_maxwell;
			} else {
				kernel_number = 410;
				mh->TxStatic = 16;
				mh->TyStatic = 16;
				mh->func_getShmPerBlk = &getShmPerBlk_GEMM_TT_rb1x1;
				mh->func_getBlkPerGrd = &getBlkPerGrd_GEMM_TT_rb1x1;
				kernel = &MUBLAS_GEMM_TT_DEVICE_tb16x16_rb1x1_s1_maxwell;
			}
		}
	}
	// ---------------------------------------------
		
	// ---------------------------------------------
	// 4. Unsupported operations (CUBLAS)
	if (kernel_number == 999) {
		#if defined (CUBLAS_GEMM)
		cublasHandle_t ch;
		cublasCreate (&ch);
		cublasSetStream (ch, mh->streamId);
		cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
		cublasOperation_t tranAcublas = CUBLAS_OP_N;
		cublasOperation_t tranBcublas = CUBLAS_OP_N;
		if (transA == 'T' || transA == 't') tranAcublas = CUBLAS_OP_T;
		if (transA == 'C' || transA == 'c') tranAcublas = CUBLAS_OP_C;
		if (transB == 'T' || transB == 't') tranBcublas = CUBLAS_OP_T;
		if (transB == 'C' || transB == 'c') tranBcublas = CUBLAS_OP_C;
		CUBLAS_GEMM (ch, tranAcublas, tranBcublas, m, n, k, alpha, a, lda, b, ldb, beta, c, ldc);
		cublasDestroy (ch);
		#else
		printf ("mublas error: not implemented");
		#endif
		mh->kernelnumber = kernel_number;
		return 0;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 5. Thread-Block Size Adjust and Kernel Launch
	cudaFuncGetAttributes (&mh->funcatb, kernel);
	cudaFuncSetCacheConfig (kernel, mh->func_cache);
	cudaDeviceSetSharedMemConfig(cudaSharedMemBankSizeEightByte);
	mublas_optimizer (mh);
	threads = dim3 (mh->Tx, mh->Ty);
	grid = dim3 (mh->Bx, mh->By);
#if defined (DBWORK)
	kernel <<< grid, threads, 0, mh->streamId >>> (m, n, k, *alpha, dev_Ab, lda_, dev_Bb, ldb_, *beta, c, ldc);
	cudaFree (dev_Ab);
	cudaFree (dev_Bb);
#else
	kernel <<< grid, threads, 0, mh->streamId >>> (m, n, k, *alpha, a, lda, b, ldb, *beta, c, ldc);
#endif
	// ---------------------------------------------

	mh->kernelnumber = kernel_number;

#ifdef MUBLAS_CC
	mublasDestroySimple (mh);
#endif

	return 0;
}

