/* ***********************************************
 
	GEMM-NN for CUDA 

   *********************************************** */   

#include "mublas_devel.h"

#define	USE_NC_A_GEMM_NN	1
#define	USE_NC_B_GEMM_NN	1

/* ***********************************************
    Macro
   *********************************************** */   
#if USE_NC_A_GEMM_NN
#define RESTRICT_A __restrict__
#define LOAD1_A LOAD1_NC
#define LOAD2_A LOAD2_NC
#define LOAD4_A LOAD4_NC
#else
#define RESTRICT_A
#define LOAD1_A LOAD1___
#define LOAD2_A LOAD2___
#define LOAD4_A LOAD4___
#endif
#if USE_NC_B_GEMM_NN
#define RESTRICT_B __restrict__
#define LOAD1_B LOAD1_NC
#define LOAD2_B LOAD2_NC
#define LOAD4_B LOAD4_NC
#else
#define RESTRICT_B
#define LOAD1_B LOAD1___
#define LOAD2_B LOAD2___
#define LOAD4_B LOAD4___
#endif
#define LOAD1_C LOAD1___
#define LOAD2_C LOAD2___
#define STORE1_C STORE1___
#define STORE2_C STORE2___

#if defined (PREC_DB) && defined (DBWORK)
#undef LOAD1___
#undef LOAD2___
#undef LOAD1_NC
#undef LOAD2_NC
#undef LOAD1_C
#undef LOAD2_C
#define LOAD1___ ld_global_f64_db
#define LOAD2___ ld_global_v2_f64_db
#define LOAD1_NC ld_global_nc_f64_db
#define LOAD2_NC ld_global_nc_v2_f64_db
#define LOAD1_C ld_global_f64
#define LOAD2_C ld_global_v2_f64
#endif

/* ***********************************************
	Functions for determining Tx and Ty
   *********************************************** */   
__host__ __inline__ void getBlkPerGrd_GEMM_NN_rb1x1 (void* mh_) {
	mublasHandle_t* mh = (mublasHandle_t*) mh_;
	const SIZE_T Bm = mh->Tx;
	const SIZE_T Bn = mh->Ty;
	mh->Bx = ceilingEx ((float)mh->m / (float)Bm, 1);
	mh->By = ceilingEx ((float)mh->n / (float)Bn, 1);
}

__host__ __inline__ void getShmPerBlk_GEMM_NN_rb1x1 (void* mh_) {
	mublasHandle_t* mh = (mublasHandle_t*) mh_;
	const SIZE_T Bm = mh->Tx;
	const SIZE_T Bn = mh->Ty;
	const SIZE_T Bk = mh->Tx;
	const SIZE_T ShmUse = Bm * Bk + Bk * Bn;
	mh->ShmPerBlk = ceilingEx ((float)(ShmUse * mh->N_one), mh->ShmUnit);
}

__host__ __inline__ void getBlkPerGrd_GEMM_NN_rb2x2 (void* mh_) {
	mublasHandle_t* mh = (mublasHandle_t*) mh_;
	const SIZE_T Bm = mh->Tx * 2;
	const SIZE_T Bn = mh->Ty * 2;
	mh->Bx = ceilingEx ((float)mh->m / (float)Bm, 1);
	mh->By = ceilingEx ((float)mh->n / (float)Bn, 1);
}

__host__ __inline__ void getShmPerBlk_GEMM_NN_rb2x2 (void* mh_) {
	mublasHandle_t* mh = (mublasHandle_t*) mh_;
	const SIZE_T Bm = mh->Tx * 2;
	const SIZE_T Bn = mh->Ty * 2;
	const SIZE_T Bk = mh->Tx;
	const SIZE_T ShmUse = Bm * Bk + Bk * Bn;
	mh->ShmPerBlk = ceilingEx ((float)(ShmUse * mh->N_one), mh->ShmUnit);
}
/*
__host__ __inline__ void getBlkPerGrd_GEMM_NN_rb4x4 (void* mh_) {
	mublasHandle_t* mh = (mublasHandle_t*) mh_;
	const SIZE_T Bm = mh->Tx * 4;
	const SIZE_T Bn = mh->Ty * 4;
	mh->Bx = ceilingEx ((float)mh->m / (float)Bm, 1);
	mh->By = ceilingEx ((float)mh->n / (float)Bn, 1);
}

__host__ __inline__ void getShmPerBlk_GEMM_NN_rb4x4 (void* mh_) {
	mublasHandle_t* mh = (mublasHandle_t*) mh_;
	const SIZE_T Bm = mh->Tx * 4;
	const SIZE_T Bn = mh->Ty * 4;
	const SIZE_T Bk = mh->Tx;
	const SIZE_T ShmUse = Bm * Bk + Bk * Bn;
	mh->ShmPerBlk = ceilingEx ((float)(ShmUse * mh->N_one), mh->ShmUnit);
}
*/
__host__ __inline__ void getBlkPerGrd_GEMM_NN_rb8x8 (void* mh_) {
	mublasHandle_t* mh = (mublasHandle_t*) mh_;
	const SIZE_T Bm = mh->Tx * 8;
	const SIZE_T Bn = mh->Ty * 8;
	mh->Bx = ceilingEx ((float)mh->m / (float)Bm, 1);
	mh->By = ceilingEx ((float)mh->n / (float)Bn, 1);
}

__host__ __inline__ void getShmPerBlk_GEMM_NN_rb8x8 (void* mh_) {
	mublasHandle_t* mh = (mublasHandle_t*) mh_;
	const SIZE_T Bm = mh->Tx * 8;
	const SIZE_T Bn = mh->Ty * 8;
	const SIZE_T Bk = mh->Tx;
	const SIZE_T ShmUse = Bm * Bk + Bk * Bn;
	mh->ShmPerBlk = ceilingEx ((float)(ShmUse * mh->N_one), mh->ShmUnit);
}

__host__ __inline__ size_t getTxMax_GEMM_NN_rb8x8 ( mublasHandle_t* mh) {
	return (size_t)ceilingEx ((float)mh->m / (float)8 / (float)mh->devprop.multiProcessorCount, mh->TxStep);
}

//__global__ void __launch_bounds__(64, 4) MUBLAS_GEMM_NN_DEVICE_tb8x8_rb8x8_v4_maxwell (
__global__ void MUBLAS_GEMM_NN_DEVICE_tb8x8_rb8x8_v4_maxwell (
	const SIZE_T m,
	const SIZE_T n,
	const SIZE_T k,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_B Bg,
	const SIZE_T ldb,
	const TYPE0 beta,
	TYPE0 *Cg,
	const SIZE_T ldc
) {
#if VEC4
	int32_t i, j, jk, ki;
	const int32_t iBx = blockIdx.x;
	const int32_t iBy = blockIdx.y;
	const int32_t iTx = threadIdx.x;
	const int32_t iTy = threadIdx.y;
	const int32_t addrx = iBx*64+iTx*4;
	const int32_t addry = iBy*64+iTy*8;
	__shared__ TYPE4 As4[8][16+1];
	__shared__ TYPE4 Bs4[8][16+1];
	register TYPE4 Tr4[8][2];
	register TYPE4 Ar4[2];
	register TYPE4 Br4[2];
	register TYPE4 Cr4[2];

	const TYPE0 * addrAg = &Ag[iTy*lda + addrx];
	const TYPE0 * addrBg = &Bg[addry*ldb + iTx];
	const TYPE0 * addrCg = &Cg[addry*ldc + addrx];

	#pragma unroll
	for (j = 0; j < 8; j++) 
		#pragma unroll
		for (i = 0; i < 2; i++)
			Tr4[j][i] = ZERO4;

	for (jk = 0; jk < k; jk += 8) {
		LOAD4_A (As4[iTy][iTx+0], addrAg+0);
		LOAD4_A (As4[iTy][iTx+8], addrAg+32);
		LOAD1_B (Br4[0].x, addrBg+0*ldb);
		LOAD1_B (Br4[0].y, addrBg+1*ldb);
		LOAD1_B (Br4[0].z, addrBg+2*ldb);
		LOAD1_B (Br4[0].w, addrBg+3*ldb);
		LOAD1_B (Br4[1].x, addrBg+4*ldb);
		LOAD1_B (Br4[1].y, addrBg+5*ldb);
		LOAD1_B (Br4[1].z, addrBg+6*ldb);
		LOAD1_B (Br4[1].w, addrBg+7*ldb);
		Bs4[iTx][iTy*2+0] = Br4[0];
		Bs4[iTx][iTy*2+1] = Br4[1];
		__syncthreads ();

		#pragma unroll
		for (ki = 0; ki < 8; ki++) {
			Ar4[0] = As4[ki][iTx+0];
			Ar4[1] = As4[ki][iTx+8];
			Br4[0] = Bs4[ki][iTy*2+0];
			Br4[1] = Bs4[ki][iTy*2+1];
			Tr4[0][0] = MAD414 (Ar4[0], Br4[0].x, Tr4[0][0]); 
			Tr4[1][0] = MAD414 (Ar4[0], Br4[0].y, Tr4[1][0]); 
			Tr4[2][0] = MAD414 (Ar4[0], Br4[0].z, Tr4[2][0]); 
			Tr4[3][0] = MAD414 (Ar4[0], Br4[0].w, Tr4[3][0]); 
			Tr4[4][0] = MAD414 (Ar4[0], Br4[1].x, Tr4[4][0]); 
			Tr4[5][0] = MAD414 (Ar4[0], Br4[1].y, Tr4[5][0]); 
			Tr4[6][0] = MAD414 (Ar4[0], Br4[1].z, Tr4[6][0]); 
			Tr4[7][0] = MAD414 (Ar4[0], Br4[1].w, Tr4[7][0]); 
			Tr4[0][1] = MAD414 (Ar4[1], Br4[0].x, Tr4[0][1]); 
			Tr4[1][1] = MAD414 (Ar4[1], Br4[0].y, Tr4[1][1]); 
			Tr4[2][1] = MAD414 (Ar4[1], Br4[0].z, Tr4[2][1]); 
			Tr4[3][1] = MAD414 (Ar4[1], Br4[0].w, Tr4[3][1]); 
			Tr4[4][1] = MAD414 (Ar4[1], Br4[1].x, Tr4[4][1]); 
			Tr4[5][1] = MAD414 (Ar4[1], Br4[1].y, Tr4[5][1]); 
			Tr4[6][1] = MAD414 (Ar4[1], Br4[1].z, Tr4[6][1]); 
			Tr4[7][1] = MAD414 (Ar4[1], Br4[1].w, Tr4[7][1]); 
		}
		__syncthreads ();
		addrAg += 8*lda;
		addrBg += 8;
	}

	#pragma unroll 
	for (j = 0; j < 8; j++) {
		Cr4[0] = (TYPE4&)*(addrCg+ 0);
		Cr4[1] = (TYPE4&)*(addrCg+32);
		Cr4[0] = MAD144 (alpha, Tr4[j][0], MUL14 (beta, Cr4[0]));
		Cr4[1] = MAD144 (alpha, Tr4[j][1], MUL14 (beta, Cr4[1]));
		(TYPE4&)*(addrCg+ 0) = Cr4[0];
		(TYPE4&)*(addrCg+32) = Cr4[1];
		addrCg += ldc;
	}

#endif
}

__global__ void MUBLAS_GEMM_NN_DEVICE_tb16x16_rb2x2_v2_maxwell (
	const SIZE_T m,
	const SIZE_T n,
	const SIZE_T k,
	const TYPE0 alpha,
#if defined (DBWORK)
	const TYPE1 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE1 * RESTRICT_A Bg,
#else
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_B Bg,
#endif
	const SIZE_T ldb,
	const TYPE0 beta,
	TYPE0 *Cg,
	const SIZE_T ldc
) {
#if VEC2
	int32_t j, jk, ki;
	const int32_t iBx = blockIdx.x;
	const int32_t iBy = blockIdx.y;
	const int32_t iTx = threadIdx.x;
	const int32_t iTy = threadIdx.y;
	const int32_t addrx = iBx*32+iTx*2;
	const int32_t addry = iBy*32+iTy*2;
	__shared__ TYPE2 As2[16][16+1];
	__shared__ TYPE2 Bs2[16][16+1];
	register TYPE2 Tr2[2];
	register TYPE2 Ar2;
	register TYPE2 Br2;
	register TYPE2 Cr2;
	Tr2[0] = ZERO2;
	Tr2[1] = ZERO2;

	for (jk = 0; jk < (k-16); jk += 16) {
		LOAD1_A (Ar2.x, &Ag[(iTy+jk)*lda + min(m-1,addrx+0)]);
		LOAD1_A (Ar2.y, &Ag[(iTy+jk)*lda + min(m-1,addrx+1)]);
		As2[iTy][iTx] = Ar2;
		LOAD1_B (Br2.x, &Bg[min(n-1,addry+0)*ldb + (iTx+jk)]);
		LOAD1_B (Br2.y, &Bg[min(n-1,addry+1)*ldb + (iTx+jk)]);
		Bs2[iTx][iTy] = Br2;
		__syncthreads ();

		#pragma unroll 
		for (ki = 0; ki < 16; ki++) {
			Ar2 = As2[ki][iTx];
			Br2 = Bs2[ki][iTy];
			Tr2[0] = MAD212 (Ar2, (TYPE1&)Br2.x, Tr2[0]); 
			Tr2[1] = MAD212 (Ar2, (TYPE1&)Br2.y, Tr2[1]); 
		}
		__syncthreads ();
	} {
		LOAD1_A (Ar2.x, &Ag[min(k-1,iTy+jk)*lda + min(m-1,addrx+0)]);
		LOAD1_A (Ar2.y, &Ag[min(k-1,iTy+jk)*lda + min(m-1,addrx+1)]);
		As2[iTy][iTx] = Ar2;
		LOAD1_B (Br2.x, &Bg[min(n-1,addry+0)*ldb + min(k-1,iTx+jk)]);
		LOAD1_B (Br2.y, &Bg[min(n-1,addry+1)*ldb + min(k-1,iTx+jk)]);
		Bs2[iTx][iTy] = Br2;
		__syncthreads ();

		for (ki = 0; ki+jk < k; ki++) {
			Ar2 = As2[ki][iTx];
			Br2 = Bs2[ki][iTy];
			Tr2[0] = MAD212 (Ar2, (TYPE1&)Br2.x, Tr2[0]); 
			Tr2[1] = MAD212 (Ar2, (TYPE1&)Br2.y, Tr2[1]); 
		}
		__syncthreads ();
	}

	TYPE0 * addrCg = &Cg[addry*ldc + addrx];
	for (j = 0; j < 2; j++) {
		if (addrx+0 < m && addry+j < n) LOAD1_C (Cr2.x, addrCg+0);
		else Cr2.x = ZERO1;
		if (addrx+1 < m && addry+j < n) LOAD1_C (Cr2.y, addrCg+1);
		else Cr2.y = ZERO1;
		Cr2 = MAD122 (LOAD0(alpha), Tr2[j], MUL12 (LOAD0(beta), Cr2));
		if (addrx+0 < m && addry+j < n) STORE1_C (addrCg+0, Cr2.x);
		if (addrx+1 < m && addry+j < n) STORE1_C (addrCg+1, Cr2.y);
		addrCg += ldc;
	}

#endif
}

__global__ void MUBLAS_GEMM_NN_DEVICE_tb16x16_rb1x1_s1_maxwell (
	const SIZE_T m,
	const SIZE_T n,
	const SIZE_T k,
	const TYPE0 alpha,
#if defined (DBWORK)
	const TYPE1 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE1 * RESTRICT_A Bg,
#else
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_B Bg,
#endif
	const SIZE_T ldb,
	const TYPE0 beta,
	TYPE0 *Cg,
	const SIZE_T ldc
) {
	int32_t jk, ki;
	const int32_t iBx = blockIdx.x;
	const int32_t iBy = blockIdx.y;
	const int32_t iTx = threadIdx.x;
	const int32_t iTy = threadIdx.y;
	const int32_t addrx = iBx*16+iTx;
	const int32_t addry = iBy*16+iTy;
	__shared__ TYPE1 As1[16][16+1];
	__shared__ TYPE1 Bs1[16][16+1];
	register TYPE1 Ar1, Br1, Cr1, Tr1;

	TYPE0 * addrCg = &Cg[addry*ldc + addrx];
	Tr1 = ZERO1;

	for (jk = 0; jk < k-16; jk += 16) {
		LOAD1_A (As1[iTy][iTx], &Ag[(jk+iTy)*lda + min(m-1,addrx)]);
		LOAD1_B (Br1, &Bg[min(n-1,addry)*ldb + (iTx+jk)]);
		Bs1[iTx][iTy] = Br1;
		__syncthreads ();
		#pragma unroll 
		for (ki = 0; ki < 16; ki++) {
			Ar1 = As1[ki][iTx];
			Br1 = Bs1[ki][iTy];
			Tr1 = MAD1 (Ar1, Br1, Tr1); 
		}
		__syncthreads ();
	} {
		LOAD1_A (As1[iTy][iTx], &Ag[min(k-1,jk+iTy)*lda + min(m-1,addrx)]);
		LOAD1_B (Br1, &Bg[min(n-1,addry)*ldb + min(k-1,iTx+jk)]);
		Bs1[iTx][iTy] = Br1;
		__syncthreads ();
		#pragma unroll 
		for (ki = 0; ki+jk < k; ki++) {
			Ar1 = As1[ki][iTx];
			Br1 = Bs1[ki][iTy];
			Tr1 = MAD1 (Ar1, Br1, Tr1); 
		}
		__syncthreads ();
	}

	if (addrx < m && addry < n) {
		LOAD1_C (Cr1, addrCg);
		Cr1 = MAD1 (LOAD0(alpha), Tr1, MUL1 (LOAD0(beta), Cr1));
		STORE1_C (addrCg, Cr1);
	}
}

#undef RESTRICT_A
#undef RESTRICT_B
#undef LOAD1_A
#undef LOAD2_A
#undef LOAD4_A
#undef LOAD1_B
#undef LOAD2_B
#undef LOAD4_B
#undef LOAD1_C
#undef LOAD2_C
#undef LOAD4_C
#undef STORE1_C
#undef STORE2_C
#undef STORE4_C
