/* ***********************************************
 
	GEMM-TN for CUDA 

   *********************************************** */   

#include "mublas_devel.h"

#define	USE_NC_A_GEMM_TN	1
#define	USE_NC_B_GEMM_TN	1

/* ***********************************************
    Macro
   *********************************************** */   
#if USE_NC_A_GEMM_TN
#define RESTRICT_A __restrict__
#define LOAD1_A LOAD1_NC
#define LOAD2_A LOAD2_NC
#define LOAD4_A LOAD4_NC
#else
#define RESTRICT_A
#define LOAD1_A LOAD1___
#define LOAD2_A LOAD2___
#define LOAD4_A LOAD4___
#endif
#if USE_NC_B_GEMM_TN
#define RESTRICT_B __restrict__
#define LOAD1_B LOAD1_NC
#define LOAD2_B LOAD2_NC
#define LOAD4_B LOAD4_NC
#else
#define RESTRICT_B
#define LOAD1_B LOAD1___
#define LOAD2_B LOAD2___
#define LOAD4_B LOAD4___
#endif
#define LOAD1_C LOAD1___
#define LOAD2_C LOAD2___
#define STORE1_C STORE1___
#define STORE2_C STORE2___

#if defined (PREC_DB) && defined (DBWORK)
#undef LOAD1___
#undef LOAD2___
#undef LOAD1_NC
#undef LOAD2_NC
#undef LOAD1_C
#undef LOAD2_C
#define LOAD1___ ld_global_f64_db
#define LOAD2___ ld_global_v2_f64_db
#define LOAD1_NC ld_global_nc_f64_db
#define LOAD2_NC ld_global_nc_v2_f64_db
#define LOAD1_C ld_global_f64
#define LOAD2_C ld_global_v2_f64
#endif

/* ***********************************************
	Functions for determining Tx and Ty
   *********************************************** */   
__host__ __inline__ void getBlkPerGrd_GEMM_TN_rb1x1 (void* mh_) {
	mublasHandle_t* mh = (mublasHandle_t*) mh_;
	const SIZE_T Bm = mh->Tx;
	const SIZE_T Bn = mh->Ty;
	mh->Bx = ceilingEx ((float)mh->m / (float)Bm, 1);
	mh->By = ceilingEx ((float)mh->n / (float)Bn, 1);
}

__host__ __inline__ void getShmPerBlk_GEMM_TN_rb1x1 (void* mh_) {
	mublasHandle_t* mh = (mublasHandle_t*) mh_;
	const SIZE_T Bm = mh->Tx;
	const SIZE_T Bn = mh->Ty;
	const SIZE_T Bk = mh->Tx;
	const SIZE_T ShmUse = Bm * Bk + Bk * Bn;
	mh->ShmPerBlk = ceilingEx ((float)(ShmUse * mh->N_one), mh->ShmUnit);
}

__host__ __inline__ void getBlkPerGrd_GEMM_TN_rb2x2 (void* mh_) {
	mublasHandle_t* mh = (mublasHandle_t*) mh_;
	const SIZE_T Bm = mh->Tx * 2;
	const SIZE_T Bn = mh->Ty * 2;
	mh->Bx = ceilingEx ((float)mh->m / (float)Bm, 1);
	mh->By = ceilingEx ((float)mh->n / (float)Bn, 1);
}

__host__ __inline__ void getShmPerBlk_GEMM_TN_rb2x2 (void* mh_) {
	mublasHandle_t* mh = (mublasHandle_t*) mh_;
	const SIZE_T Bm = mh->Tx * 2;
	const SIZE_T Bn = mh->Ty * 2;
	const SIZE_T Bk = mh->Tx;
	const SIZE_T ShmUse = Bm * Bk + Bk * Bn;
	mh->ShmPerBlk = ceilingEx ((float)(ShmUse * mh->N_one), mh->ShmUnit);
}

__host__ __inline__ size_t getTxMax_GEMM_TN_rb8x8 ( mublasHandle_t* mh) {
	return (size_t)ceilingEx ((float)mh->m / (float)8 / (float)mh->devprop.multiProcessorCount, mh->TxStep);
}

__global__ void MUBLAS_GEMM_TN_DEVICE_tb16x16_rb1x1_s1_maxwell (
	const SIZE_T m, const SIZE_T n, const SIZE_T k,
	const TYPE0 alpha,
#if defined (DBWORK)
	const TYPE1 * RESTRICT_A Ag, const SIZE_T lda, const TYPE1 * RESTRICT_A Bg,
#else
	const TYPE0 * RESTRICT_A Ag, const SIZE_T lda, const TYPE0 * RESTRICT_B Bg,
#endif
	const SIZE_T ldb,
	const TYPE0 beta, TYPE0 *Cg, const SIZE_T ldc
) {
	int32_t jk, ki;
	const int32_t iBx = blockIdx.x;
	const int32_t iBy = blockIdx.y;
	const int32_t iTx = threadIdx.x;
	const int32_t iTy = threadIdx.y;
	const int32_t addrx = iBx*16+iTx;
	const int32_t addry = iBy*16+iTy;
	__shared__ TYPE1 As1[16][16+1];
	__shared__ TYPE1 Bs1[16][16+1];
	register TYPE1 Ar1, Br1, Cr1, Tr1;
	TYPE0 * addrCg = &Cg[addry*ldc + addrx];
	Tr1 = ZERO1;

	for (jk = 0; jk < k-16; jk += 16) {
		LOAD1_A (Ar1, &Ag[min(m-1,addrx)*lda + (iTy+jk)]);
		As1[iTy][iTx] = Ar1;
		LOAD1_B (Br1, &Bg[min(n-1,addry)*ldb + (iTx+jk)]);
		Bs1[iTx][iTy] = Br1;
		__syncthreads ();

		#pragma unroll 
		for (ki = 0; ki < 16; ki++) {
			Ar1 = As1[ki][iTx];
			Br1 = Bs1[ki][iTy];
			Tr1 = MAD1 (Ar1, Br1, Tr1); 
		}
		__syncthreads ();
	}
	{
		LOAD1_A (Ar1, &Ag[min(m-1,addrx)*lda + min(k-1,iTy+jk)]);
		As1[iTy][iTx] = Ar1;
		LOAD1_B (Br1, &Bg[min(n-1,addry)*ldb + min(k-1,iTx+jk)]);
		Bs1[iTx][iTy] = Br1;
		__syncthreads ();

		for (ki = 0; ki+jk < k; ki++) {
			Ar1 = As1[ki][iTx];
			Br1 = Bs1[ki][iTy];
			Tr1 = MAD1 (Ar1, Br1, Tr1); 
		}
		__syncthreads ();
	}

	if (addrx < m && addry < n) {
		LOAD1_C (Cr1, addrCg);
		Cr1 = MAD1 (LOAD0(alpha), Tr1, MUL1 (LOAD0(beta), Cr1));
		STORE1_C (addrCg, Cr1);
	}
}



__global__ void __launch_bounds__(256, 6) MUBLAS_GEMM_TN_DEVICE_tb16x16_rb2x2_v2_maxwell (
	const SIZE_T m,
	const SIZE_T n,
	const SIZE_T k,
	const TYPE0 alpha,
#if defined (DBWORK)
	const TYPE1 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE1 * RESTRICT_A Bg,
#else
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_B Bg,
#endif
	const SIZE_T ldb,
	const TYPE0 beta,
	TYPE0 *Cg,
	const SIZE_T ldc
) {
#if VEC2
	int32_t j, jk, ki;
	const int32_t iBx = blockIdx.x;
	const int32_t iBy = blockIdx.y;
	const int32_t iTx = threadIdx.x;
	const int32_t iTy = threadIdx.y;
	const int32_t addrx = iBx*32+iTx*2;
	const int32_t addry = iBy*32+iTy*2;
	__shared__ TYPE2 As2[16][16+1];
	__shared__ TYPE2 Bs2[16][16+1];
	register TYPE2 Tr2[2];
	register TYPE2 Ar2;
	register TYPE2 Br2;
	register TYPE2 Cr2;
	Tr2[0] = ZERO2;
	Tr2[1] = ZERO2;

	for (jk = 0; jk < (k-16); jk += 16) {
		LOAD1_A (Ar2.x, &Ag[min(m-1,addrx+0)*lda + (iTy+jk)]);
		LOAD1_A (Ar2.y, &Ag[min(m-1,addrx+1)*lda + (iTy+jk)]);
		As2[iTy][iTx] = Ar2;
		LOAD1_B (Br2.x, &Bg[min(n-1,addry+0)*ldb + (iTx+jk)]);
		LOAD1_B (Br2.y, &Bg[min(n-1,addry+1)*ldb + (iTx+jk)]);
		Bs2[iTx][iTy] = Br2;
		__syncthreads ();

		#pragma unroll 
		for (ki = 0; ki < 16; ki++) {
			Ar2 = As2[ki][iTx];
			Br2 = Bs2[ki][iTy];
			Tr2[0] = MAD212 (Ar2, (TYPE1&)Br2.x, Tr2[0]); 
			Tr2[1] = MAD212 (Ar2, (TYPE1&)Br2.y, Tr2[1]); 
		}
		__syncthreads ();
	} {
		LOAD1_A (Ar2.x, &Ag[min(m-1,addrx+0)*lda + min(k-1,iTy+jk)]);
		LOAD1_A (Ar2.y, &Ag[min(m-1,addrx+1)*lda + min(k-1,iTy+jk)]);
		As2[iTy][iTx] = Ar2;
		LOAD1_B (Br2.x, &Bg[min(n-1,addry+0)*ldb + min(k-1,iTx+jk)]);
		LOAD1_B (Br2.y, &Bg[min(n-1,addry+1)*ldb + min(k-1,iTx+jk)]);
		Bs2[iTx][iTy] = Br2;
		__syncthreads ();

		for (ki = 0; ki+jk < k; ki++) {
			Ar2 = As2[ki][iTx];
			Br2 = Bs2[ki][iTy];
			Tr2[0] = MAD212 (Ar2, (TYPE1&)Br2.x, Tr2[0]); 
			Tr2[1] = MAD212 (Ar2, (TYPE1&)Br2.y, Tr2[1]); 
		}
		__syncthreads ();
	}

	TYPE0 * addrCg = &Cg[addry*ldc + addrx];
	for (j = 0; j < 2; j++) {
		if (addrx+0 < m && addry+j < n) LOAD1_C (Cr2.x, addrCg+0);
		else Cr2.x = ZERO1;
		if (addrx+1 < m && addry+j < n) LOAD1_C (Cr2.y, addrCg+1);
		else Cr2.y = ZERO1;
		Cr2 = MAD122 (LOAD0(alpha), Tr2[j], MUL12 (LOAD0(beta), Cr2));
		if (addrx+0 < m && addry+j < n) STORE1_C (addrCg+0, Cr2.x);
		if (addrx+1 < m && addry+j < n) STORE1_C (addrCg+1, Cr2.y);
		addrCg += ldc;
	}

#endif
}

#undef RESTRICT_A
#undef RESTRICT_B
#undef LOAD1_A
#undef LOAD2_A
#undef LOAD4_A
#undef LOAD1_B
#undef LOAD2_B
#undef LOAD4_B
#undef LOAD1_C
#undef LOAD2_C
#undef LOAD4_C
#undef STORE1_C
#undef STORE2_C
#undef STORE4_C
