/* ***********************************************
 
	GEMV for CUDA 

   *********************************************** */   

#include "mublas_devel.h"
#include "mublas_gemv_mgpu_n_kernel.cu"
#include "mublas_gemv_mgpu_t_kernel.cu"

/* ***********************************************
	Host Interface
   *********************************************** */   

__host__ int32_t
MUBLAS_GEMV_MGPU_ASYNC (
	mublasHandle_t *mh,	
	const char trans,
	const size_t m,
	const size_t n,
	const TYPE0 alpha,
	TYPE0 **a,
	const size_t lda,
	TYPE0 **x,
	const size_t incx,
	const TYPE0 beta,
	TYPE0 **y,
	const size_t incy,
	const uint32_t ngpus,
	const size_t bs
) {
	uint32_t kernel_number = 999;
	uint32_t nonvectorkernel = 0;
	void (*kernel) (const SIZE_T, const SIZE_T, const TYPE0, const TYPE0*, const SIZE_T, const TYPE0*, const TYPE0, TYPE0*, const uint32_t, const uint32_t, const SIZE_T);
	int64_t tt;
	uint32_t igpu;
	size_t nr, ni[NGPU_MAX] = {0};
	mublasRecipeDefault (&mh[0]);

	// ---------------------------------------------
	// 1. Shortcut
	if (m == 0 || n == 0 || (alpha == ZERO0 && beta == ONE0 )) 
		return 0;
	/*
	if (alpha == ZERO0) {
		MUBLAS_SCAL (mh, m, &beta, y, incy);
		return 0;
	}
	*/
	// ---------------------------------------------

	// ---------------------------------------------
	// 2. Kernel Configuration 
	mh[0].N_one = sizeof (TYPE0);
	mh[0].N_T = 1;
	if      (VEC4) mh->N_T = 4;
	else if (VEC2) mh->N_T = 2;
	if ((((long)a%(128/sizeof(TYPE0)))%16 != 0) || 
		(((long)x%(128/sizeof(TYPE0)))%16 != 0) ||
		(((long)y%(128/sizeof(TYPE0)))%16 != 0) ||
		(lda%mh->N_T != 0)) nonvectorkernel = 1;
	if (incx == 1 && incy == 1) {
		if (trans == 'N' || trans == 'n') {
			kernel_number = 100 + mh->N_T * 10 + nonvectorkernel;
			mh[0].BlkDim = 2;
			mh[0].func_getShmPerBlk = &getShmPerBlk_GEMV_N;
			mh[0].func_getBlkPerGrd = &getBlkPerGrd_GEMV_N;
			mh[0].func_cache = cudaFuncCachePreferShared;
			#include "mublas_gemv_n.incl"
		} else if (trans == 'T' || trans == 't') {
			kernel_number = 200 + mh->N_T * 10 + nonvectorkernel;
			mh[0].BlkDim = 1;
			mh[0].func_getShmPerBlk = &getShmPerBlk_GEMV_T;
			mh[0].func_getBlkPerGrd = &getBlkPerGrd_GEMV_T;
			mh[0].func_cache = cudaFuncCachePreferL1;
			#include "mublas_gemv_t.incl"
		}
	}
	/*
	if (mh->kerneldebug != 0) {
		kernel_number = mh->kerneldebug;
		switch (kernel_number) {
			case 110: mh->N_T = 1; break;
			case 120: mh->N_T = 2; break;
			case 140: mh->N_T = 4; break;
			case 210: mh->N_T = 1; break;
			case 220: mh->N_T = 2; break;
			case 240: mh->N_T = 4; break;
		}
	}
	*/
	switch (kernel_number) {
		case 140: kernel = &MUBLAS_GEMV_MGPU_N_DEVICE_v4; break;
		case 141: kernel = &MUBLAS_GEMV_MGPU_N_DEVICE_s4; break;
		case 120: kernel = &MUBLAS_GEMV_MGPU_N_DEVICE_v2; break;
		case 121: kernel = &MUBLAS_GEMV_MGPU_N_DEVICE_s2; break;
		case 110: kernel = &MUBLAS_GEMV_MGPU_N_DEVICE_s1; break;
		case 240: kernel = &MUBLAS_GEMV_MGPU_T_DEVICE_v4; break;
		case 220: kernel = &MUBLAS_GEMV_MGPU_T_DEVICE_v2; break;
		case 210: kernel = &MUBLAS_GEMV_MGPU_T_DEVICE_s1; break;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 3. Unsupported operations (CUBLAS)
	if (kernel_number == 999) {
		printf ("mublas error: not implemented");
		mh->kernelnumber = kernel_number;
		return 0;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 4. set m and n dims
	if (trans == 'N' || trans == 'n') {
		for (igpu = 0; igpu < ngpus; igpu++) {
			mh[igpu].m = m;
			tt = (int64_t)((n/bs)%ngpus)-(int64_t)igpu;
			mh[igpu].n = bs*((n/bs)/ngpus);
			mh[igpu].n += (tt>0)?bs:(tt==0)?n%bs:0;
		}
	} else {
		nr = n;
		igpu = ngpus;
		while (nr >= bs) {
			ni[igpu%ngpus] += bs;
			nr -= bs;
			igpu++;
		} 
		ni[igpu%ngpus] += nr;
		mh[0].m = m;
		mh[0].n = ni[0];
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 5. Thread-Block Size Adjustment and Kernel Launch
	cudaFuncGetAttributes (&mh[0].funcatb, kernel);
	cudaFuncSetCacheConfig (kernel, mh[0].func_cache);
	mublas_optimizer (&mh[0]);
	dim3 threads = dim3 (mh[0].Tx, mh[0].Ty);
	dim3 grid = dim3 (mh[0].Bx, mh[0].By);

	mublasHandle_t mh_scal;
	TYPE0 betai;
	if (trans == 'N' || trans == 'n') {
		for (igpu = 0; igpu < ngpus; igpu++) {
			cudaSetDevice (igpu);
			betai = ZERO0;
			if (mh[igpu].n == 0) {
				mh_scal = mh[igpu];
				MUBLAS_SCAL (&mh_scal, m, &betai, y[igpu], 1);
			} else {
				if (igpu == 0) betai = beta;
				kernel <<< grid, threads, mh[0].ShmPerBlk, mh[igpu].streamId >>> (m, mh[igpu].n, alpha, a[igpu], lda, x[igpu], betai, y[igpu], ngpus, igpu, bs);
			}
		}
	} else {
		for (igpu = 0; igpu < ngpus; igpu++) {
			cudaSetDevice (igpu);
			betai = ZERO0;
			if (igpu == 0) betai = beta;
			mh_scal = mh[igpu];
			MUBLAS_SCAL (&mh_scal, n, &betai, y[igpu], 1);
			if (igpu == 0) betai = ONE0;
			mh[igpu].n = ni[igpu]; // mh[igpu].n is reset on MUBLAS_SCAL
			if (mh[igpu].n != 0) 
				kernel <<< grid, threads, mh[0].ShmPerBlk, mh[igpu].streamId >>> (m, mh[igpu].n, alpha, a[igpu], lda, x[igpu], betai, y[igpu], ngpus, igpu, bs);
		}
	}
	// ---------------------------------------------

	mh->kernelnumber = kernel_number;
	return 0;
}

