/* ***********************************************

SCAL

memo:
Actually v2&v4 kernels are not needed.
v1 kernel with AT is enought to achieve best throughput.

  *********************************************** */   

#include "mublas_scal.def"

/* ***********************************************
    Macro
   *********************************************** */   
#define LOAD1_X LOAD1___
#define LOAD2_X LOAD2___
#define LOAD4_X LOAD4___
#define STORE1_X STORE1___
#define STORE2_X STORE2___
#define STORE4_X STORE4___

/* ***********************************************
	Functions for determining Tx 
   *********************************************** */   
__host__ __inline__ void getBlkPerGrd_SCAL (void* mh_) {
	mublasHandle_t* mh = (mublasHandle_t*) mh_;
	mh->Bx = ceilingEx ((float)mh->n / (float)(mh->N_T * mh->Tx), 1);
	mh->By = 1;
}

__host__ __inline__ void getShmPerBlk_SCAL (void* mh_) {
	mublasHandle_t* mh = (mublasHandle_t*) mh_;
	mh->ShmPerBlk = 0.;
}

__host__ __inline__ size_t getTxSuf_SCAL (mublasHandle_t* mh) {
	return (size_t)ceilingEx ((float)mh->n / (float)mh->N_T / (float)mh->devprop.multiProcessorCount, mh->TxStep);
}

/* ***********************************************
	kernel=v4
   *********************************************** */   

__global__ void MUBLAS_SCAL_DEVICE_v4 (
	const SIZE_T n,
	const TYPE0 alpha,
	TYPE0 *x
) {
#if VEC4
	const SIZE_T N_T = 4;
	const SIZE_T Tx = blockDim.x;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T rowid = blockIdx.x * Tx * N_T + iTx * N_T;
	TYPE4 Xr4;

	if (rowid < n) {
		LOAD4_X (Xr4, &x[rowid]);
		Xr4 = MUL14 (LOAD0(alpha), Xr4);
		STORE4_X (&x[rowid], Xr4);
	}
#endif
}

/* ***********************************************
	kernel=v2
   *********************************************** */   

__global__ void MUBLAS_SCAL_DEVICE_v2 (
	const SIZE_T n,
	const TYPE0 alpha,
	TYPE0 *x
) {
#if VEC2
	const SIZE_T N_T = 2;
	const SIZE_T Tx = blockDim.x;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T rowid = blockIdx.x * Tx * N_T + iTx * N_T;
	TYPE2 Xr2;

	if (rowid < n) {
		LOAD2_X (Xr2, &x[rowid]);
		Xr2 = MUL12 (LOAD0(alpha), Xr2);
		STORE2_X (&x[rowid], Xr2);
	}
#endif
}


/* ***********************************************
	kernel=s1
   *********************************************** */   

__global__ void MUBLAS_SCAL_DEVICE_s1 (
	const SIZE_T n,
	const TYPE0 alpha,
	TYPE0 *x
) {
#if VEC1
	const SIZE_T Tx = blockDim.x;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T rowid = blockIdx.x * Tx + iTx;
	TYPE1 Xr1;

	if (rowid < n) {
		LOAD1_X (Xr1, &x[rowid]);
		Xr1 = MUL1 (LOAD0(alpha), Xr1);
		STORE1_X (&x[rowid], Xr1);
	}

#endif
}

#undef LOAD1_X
#undef LOAD2_X
#undef LOAD4_X
#undef STORE1_X
#undef STORE2_X
#undef STORE4_X

