#ifndef PREC_U_H
#define PREC_U_H

#include "cudd.h"

#ifdef PREC_U
	// host
	#define MUBLAS_GEMV_MGPU_ASYNC mublasUgemvMgpuAsync
	#define MUBLAS_GEMV_MGPU mublasUgemvMgpu
	#define MUBLAS_CSRMV mublasUcsrmv
	#define MUBLAS_GEAM mublasUgeam
	#define MUBLAS_GEMM mublasUgemm
	#define MUBLAS_GEMV mublasUgemv
	#define MUBLAS_HEMV mublasUhemv
	#define MUBLAS_TRMV mublasUtrmv
	#define MUBLAS_AXPY mublasUaxpy
	#define MUBLAS_DOT  mublasUdot
	#define MUBLAS_SCAL mublasUscal
	#define MUBLAS_COPY mublasUcopy
//	#define CUBLAS_GEMM 
//	#define CUBLAS_GEMV 
//	#define CUBLAS_HEMV 
//	#define CUBLAS_TRMV 
//	#define CUBLAS_AXPY 
//	#define CUBLAS_DOT
//	#define CUBLAS_SCAL 
//	#define CUBLAS_COPY 
	#define TYPE0 cuddcomplex
	#define ZERO0 make_cuddcomplex(make_cuddreal(0.,0.),make_cuddreal(0.,0.))
	#define ONE0 make_cuddcomplex(make_cuddreal(1.,0.),make_cuddreal(0.,0.))
	#define LOAD0(A) (A)

	// vec1
	#define VEC1 1
	#define TYPE1 TYPE0
	#define ZERO1 ZERO0
	#define ADD1 cuUadd
	#define MUL1 cuUmul
	#define MAD1 cuUmad
	#define LOAD1___ ld_global_v2_f64_x2
	#define LOAD1_NC ld_global_nc_v2_f64_x2
	#define STORE1___ st_global_v2_f64_x2

	// vec2
	#define VEC2 0

	// vec4
	#define VEC4 0

	// DD-algorithm 
	#define CUWADD cuWaddV2
	#define CUWSUB cuWsubV2
	#define CUWMUL cuWmulV3
	#define CUWMAD cuWmuladd
#endif

__host__ __device__ __forceinline__ bool operator== (const cuddreal a, const cuddreal b) {
	if ((a.x == b.x) && (a.y == b.y))
		return true;
	else
		return false;
}

__host__ __device__ __forceinline__ bool operator== (const cuddcomplex a, const cuddcomplex b) {
	if ((a.x == b.x) && (a.y == b.y))
		return true;
	else
		return false;
}

__device__ __forceinline__ cuddcomplex
cuUadd
	(const cuddcomplex a, const cuddcomplex b)
{
	cuddcomplex c;
	c.x = CUWADD (a.x, b.x);
	c.y = CUWADD (a.y, b.y);
	return c;
}

__device__ __forceinline__ cuddcomplex
cuUmul
	(const cuddcomplex a, const cuddcomplex b)
{
	cuddcomplex c;
	c.x = CUWSUB (CUWMUL (a.x, b.x), CUWMUL (a.y, b.y));
	c.y = CUWADD (CUWMUL (a.x, b.y), CUWMUL (a.y, b.x));
	return c;
}

__device__ __forceinline__ cuddcomplex
cuUmad
	(const cuddcomplex a, const cuddcomplex b, const cuddcomplex c)
{
	cuddcomplex d;
	d.x = CUWADD( CUWSUB (CUWMUL (a.x, b.x), CUWMUL (a.y, b.y)), c.x);
	d.y = CUWADD( CUWADD (CUWMUL (a.x, b.y), CUWMUL (a.y, b.x)), c.y);
	return d;
}

// === Shuffle =====================================================
__device__ __forceinline__ void shuffleAdd (cuddcomplex &val, const int n) {
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	val = cuUadd (val, make_cuddcomplex ( 
		make_cuddreal(
		__hiloint2double( __shfl_xor(__double2hiint(val.x.x), n, WrpSize), __shfl_xor(__double2loint(val.x.x), n, WrpSize)),
		__hiloint2double( __shfl_xor(__double2hiint(val.x.y), n, WrpSize), __shfl_xor(__double2loint(val.x.y), n, WrpSize))),
		make_cuddreal(
		__hiloint2double( __shfl_xor(__double2hiint(val.y.x), n, WrpSize), __shfl_xor(__double2loint(val.y.x), n, WrpSize)),
		__hiloint2double( __shfl_xor(__double2hiint(val.y.y), n, WrpSize), __shfl_xor(__double2loint(val.y.y), n, WrpSize)))));
#endif
}
// =================================================================

// === Load ========================================================
__device__ __forceinline__ void ld_global_nc_v2_f64_x2 (cuddcomplex &reg, const cuddcomplex* addr) {
#if USE_INLINE_ASM
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.v2.f64 {%0,%1}, [%2+0];": "=d"(reg.x.x), "=d"(reg.x.y): "l"(addr));
	asm ("ld.global.nc.v2.f64 {%0,%1}, [%2+16];": "=d"(reg.y.x), "=d"(reg.y.y): "l"(addr));
	#else
	asm ("ld.global.v2.f64 {%0,%1}, [%2+0];": "=d"(reg.x.x), "=d"(reg.x.y): "l"(addr));
	asm ("ld.global.v2.f64 {%0,%1}, [%2+16];": "=d"(reg.y.x), "=d"(reg.y.y): "l"(addr));
	#endif
#else
	reg = *addr;
#endif
}

__device__ __forceinline__ void ld_global_v2_f64_x2 (cuddcomplex &reg, const cuddcomplex* addr) {
#if USE_INLINE_ASM
	asm ("ld.global.v2.f64 {%0,%1}, [%2+0];": "=d"(reg.x.x), "=d"(reg.x.y): "l"(addr));
	asm ("ld.global.v2.f64 {%0,%1}, [%2+16];": "=d"(reg.y.x), "=d"(reg.y.y): "l"(addr));
#else
	reg = *addr;
#endif
}
// =================================================================

// === Store =======================================================
__device__ __forceinline__ void st_global_v2_f64_x2 (cuddcomplex *addr, const cuddcomplex reg) {
#if USE_INLINE_ASM
	asm ("st.global.v2.f64 [%0], {%1,%2};":: "l"(addr), "d"(reg.x.x), "d"(reg.x.y));
	asm ("st.global.v2.f64 [%0+16], {%1,%2};":: "l"(addr), "d"(reg.y.x), "d"(reg.y.y));
#else
	*addr = reg;
#endif
}
// =================================================================

#endif
