#ifndef PREC_V_H
#define PREC_V_H

#include "cudf.h"

#ifdef PREC_V
	// host
	#define MUBLAS_GEMV_MGPU_ASYNC mublasVgemvMgpuAsync
	#define MUBLAS_GEMV_MGPU mublasVgemvMgpu
	#define MUBLAS_CSRMV mublasVcsrmv
	#define MUBLAS_GEAM mublasVgeam
	#define MUBLAS_GEMM mublasVgemm
	#define MUBLAS_GEMV mublasVgemv
	#define MUBLAS_HEMV mublasVhemv
	#define MUBLAS_TRMV mublasVtrmv
	#define MUBLAS_AXPY mublasVaxpy
	#define MUBLAS_DOT mublasVdot
	#define MUBLAS_SCAL mublasVscal
	#define MUBLAS_COPY mublasVcopy
//	#define CUBLAS_GEMM 
//	#define CUBLAS_GEMV 
//	#define CUBLAS_HEMV 
//	#define CUBLAS_TRMV 
//	#define CUBLAS_AXPY 
//	#define CUBLAS_DOT
//	#define CUBLAS_SCAL 
//	#define CUBLAS_COPY 
	#define TYPE0 cudfcomplex
	#define ZERO0 make_cudfcomplex(make_cudfreal(0.,0.),make_cudfreal(0.,0.))
	#define ONE0 make_cudfcomplex(make_cudfreal(1.,0.),make_cudfreal(0.,0.))
	#define LOAD0(A) (A)

	// vec1
	#define VEC1 1
	#define TYPE1 TYPE0
	#define ZERO1 ZERO0
	#define ADD1 cuVadd
	#define MUL1 cuVmul
	#define MAD1 cuVmad
	#define LOAD1___ ld_global_v4_f32
	#define LOAD1_NC ld_global_nc_v4_f32
	#define STORE1___ st_global_v4_f32

	// vec2
	#define VEC2 0

	// vec4
	#define VEC4 0

	// DF-algorithm 
	#define CUBADD cuBaddV2
	#define CUBSUB cuBsubV2
	#define CUBMUL cuBmulV3
	#define CUBMAD cuBmuladd
#endif



__host__ __device__ __forceinline__ bool operator== (const cudfreal a, const cudfreal b) {
	if ((a.x == b.x) && (a.y == b.y))
		return true;
	else
		return false;
}

__host__ __device__ __forceinline__ bool operator== (const cudfcomplex a, const cudfcomplex b) {
	if ((a.x == b.x) && (a.y == b.y))
		return true;
	else
		return false;
}

__device__ __forceinline__ cudfcomplex
cuVadd
	(const cudfcomplex a, const cudfcomplex b)
{
	cudfcomplex c;
	c.x = CUBADD (a.x, b.x);
	c.y = CUBADD (a.y, b.y);
	return c;
}

__device__ __forceinline__ cudfcomplex
cuVmul
	(const cudfcomplex a, const cudfcomplex b)
{
	cudfcomplex c;
	c.x = CUBSUB (CUBMUL (a.x, b.x), CUBMUL (a.y, b.y));
	c.y = CUBADD (CUBMUL (a.x, b.y), CUBMUL (a.y, b.x));
	return c;
}

__device__ __forceinline__ cudfcomplex
cuVmad
	(const cudfcomplex a, const cudfcomplex b, const cudfcomplex c)
{
	cudfcomplex d;
	d.x = CUBADD( CUBSUB (CUBMUL (a.x, b.x), CUBMUL (a.y, b.y)), c.x);
	d.y = CUBADD( CUBADD (CUBMUL (a.x, b.y), CUBMUL (a.y, b.x)), c.y);
	return d;
}

// === Shuffle =====================================================
__device__ __forceinline__ void shuffleAdd (cudfcomplex &val, const int n) {
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	val = cuVadd (val, make_cudfcomplex ( 
		make_cudfreal (__shfl_xor(val.x.x, n, WrpSize), __shfl_xor(val.x.y, n, WrpSize)),
		make_cudfreal (__shfl_xor(val.y.x, n, WrpSize), __shfl_xor(val.y.y, n, WrpSize))));
#endif
}
// =================================================================

// === Load ========================================================
__device__ __forceinline__ void ld_global_nc_v4_f32 (cudfcomplex &reg, const cudfcomplex *addr) {
#if USE_INLINE_ASM
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.v4.f32 {%0,%1,%2,%3}, [%4];": "=f"(reg.x.x), "=f"(reg.x.y), "=f"(reg.y.x), "=f"(reg.y.y): "l"(addr));
	#else
	asm ("ld.global.v4.f32 {%0,%1,%2,%3}, [%4];": "=f"(reg.x.x), "=f"(reg.x.y), "=f"(reg.y.x), "=f"(reg.y.y): "l"(addr));
	#endif
#else
	reg = *addr;
#endif
}

__device__ __forceinline__ void ld_global_v4_f32 (cudfcomplex &reg, const cudfcomplex *addr) {
#if USE_INLINE_ASM
	asm ("ld.global.v4.f32 {%0,%1,%2,%3}, [%4];": "=f"(reg.x.x), "=f"(reg.x.y), "=f"(reg.y.x), "=f"(reg.y.y): "l"(addr));
#else
	reg = *addr;
#endif
}
// =================================================================

// === Store =======================================================
__device__ __forceinline__ void st_global_v4_f32 (cudfcomplex *addr, const cudfcomplex reg) {
#if USE_INLINE_ASM
	asm ("st.global.v4.f32 [%0], {%1,%2,%3,%4};":: "l"(addr), "f"(reg.x.x), "f"(reg.x.y), "f"(reg.y.x), "f"(reg.y.y));
#else
	(cudfcomplex&)*addr = reg;
#endif
}
// =================================================================




#endif
