#include "testing_common.h"
#include "testing_level3.h"
#include "testing_common.cpp"


int32_t
main (int32_t argc, char **argv)
{
	size_t i, iter = MINLOOP;
	size_t lda_dev, lda_hst, rda_dev, rda_hst, cda_dev, cda_hst;
	size_t ldb_dev, ldb_hst, rdb_dev, rdb_hst, cdb_dev, cdb_hst;
	size_t ldc_dev, ldc_hst, rdc_dev, rdc_hst, cdc_dev, cdc_hst;
	size_t word_size_r = sizeof (REFERENCE_TYPE);
	double t0, t1;
	cudaError_t err;

// testing setup ------------------------------
	testingHandle_t th;
	testingCreate (argc, argv, &th);
	print_info1 (&th);
// --------------------------------------------

// library setup ------------------------------
	#if defined (CHECK_CUBLAS)
	cublasHandle_t ch;
	cublasCreate (&ch);
	cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
	#endif
	#if defined (MUBLAS)
	mublasHandle_t ha;
	mublasCreate (&ha);
	set_mublas_handler (&th, &ha);
	#elif defined (CUBLAS)
	cublasHandle_t ha;
	cublasCreate (&ha);
	cublasSetPointerMode(ha, CUBLAS_POINTER_MODE_HOST);
	if (th.cublasatomic == 1) 
		cublasSetAtomicsMode(ha, CUBLAS_ATOMICS_ALLOWED);
	else
		cublasSetAtomicsMode(ha, CUBLAS_ATOMICS_NOT_ALLOWED);
	#elif defined (KBLAS)
	#elif defined (MAGMA)
	magma_init();
	#elif defined (CULA)
    culaInitialize();
	#elif defined (ASPEN)
	ASPEN_init(0);
	#else
	#endif
#if DEBUG_ALLSKIP == 0
// --------------------------------------------

// memory setup -------------------------------
	GPU_TYPE alpha, beta;
	GPU_TYPE *hst_A, *hst_B, *hst_C, *hst_C_t;
	GPU_TYPE *dev_A, *dev_B, *dev_C;
	REFERENCE_TYPE alpha_r = REFERENCE_ZERO, beta_r = REFERENCE_ZERO;
	REFERENCE_TYPE *hst_A_r = 0, *hst_B_r = 0, *hst_C_r = 0;

	// dim setup
	dim_hst_setup (&th);
	// A
	rda_hst = th.dim_stop;
	cda_hst = th.dim_stop;
	if (th.tranA == 'N' || th.tranA == 'n') {
		if (th.dim_m_const > 0) rda_hst = th.dim_m_const;
		if (th.dim_k_const > 0) cda_hst = th.dim_k_const;
	} else {
		if (th.dim_k_const > 0) rda_hst = th.dim_k_const;
		if (th.dim_m_const > 0) cda_hst = th.dim_m_const;
	}
	lda_hst = rda_hst;
	// B
	rdb_hst = th.dim_stop;
	cdb_hst = th.dim_stop;
	if (th.tranB == 'N' || th.tranB == 'n') {
		if (th.dim_k_const > 0) rdb_hst = th.dim_k_const;
		if (th.dim_n_const > 0) cdb_hst = th.dim_n_const;
	} else {
		if (th.dim_n_const > 0) rdb_hst = th.dim_n_const;
		if (th.dim_k_const > 0) cdb_hst = th.dim_k_const;
	}
	ldb_hst = rdb_hst;
	// C
	rdc_hst = th.dim_stop;
	cdc_hst = th.dim_stop;
	if (th.dim_m_const > 0) rdc_hst = th.dim_m_const;
	if (th.dim_n_const > 0) cdc_hst = th.dim_n_const;
	ldc_hst = rdc_hst;
	// malloc host memory
	cudaMallocHost ((void **) &hst_A, th.word_size * lda_hst * cda_hst);
	cudaMallocHost ((void **) &hst_B, th.word_size * ldb_hst * cdb_hst);
	cudaMallocHost ((void **) &hst_C, th.word_size * ldc_hst * cdc_hst);
	// initialize
	mublasInit2D (1, 1, 0, &alpha);
	mublasInit2D (1, 1, 0, &beta);
	mublasInit2D (lda_hst, cda_hst, lda_hst, hst_A);
	mublasInit2D (ldb_hst, cdb_hst, ldb_hst, hst_B);
	mublasInit2D (ldc_hst, cdc_hst, ldc_hst, hst_C);
	if (th.mode == 'c') {
		// initialize
		cudaMallocHost ((void **) &hst_A_r, word_size_r * lda_hst * cda_hst);
		cudaMallocHost ((void **) &hst_B_r, word_size_r * ldb_hst * cdb_hst);
		cudaMallocHost ((void **) &hst_C_r, word_size_r * ldc_hst * cdc_hst);
		cudaMallocHost ((void **) &hst_C_t, th.word_size * ldc_hst * cdc_hst);
	}
// --------------------------------------------

	print_info2 (&th);

// evaluation ---------------------------------
	dim_dev_setup (&th);
	while (1) {
		if ((th.dim_m_const == 0 && th.dim_m_dev > th.dim_stop) ||
		    (th.dim_n_const == 0 && th.dim_n_dev > th.dim_stop) ||
		    (th.dim_k_const == 0 && th.dim_k_dev > th.dim_stop)) break;
		// dim setup
		// A
		if (th.tranA == 'N' || th.tranA == 'n') {
			rda_dev = th.dim_m_dev;
			cda_dev = th.dim_k_dev;
		} else {
			rda_dev = th.dim_k_dev;
			cda_dev = th.dim_m_dev;
		}
		// B
		if (th.tranB == 'N' || th.tranB == 'n') {
			rdb_dev = th.dim_k_dev;
			cdb_dev = th.dim_n_dev;
		} else {
			rdb_dev = th.dim_n_dev;
			cdb_dev = th.dim_k_dev;
		}
		// C
		rdc_dev = th.dim_m_dev;
		cdc_dev = th.dim_n_dev;
		// malloc device memory
		#if PITCHALLOC
		size_t pitch;
		cudaMallocPitch ((void **) &dev_A, &pitch, th.word_size * rda_dev, cda_dev);
		lda_dev = pitch/th.word_size;
		cudaMallocPitch ((void **) &dev_B, &pitch, th.word_size * rdb_dev, cdb_dev);
		ldb_dev = pitch/th.word_size;
		cudaMallocPitch ((void **) &dev_C, &pitch, th.word_size * rdc_dev, cdc_dev);
		ldc_dev = pitch/th.word_size;
		#else
		lda_dev = (th.lda_dev!=0) ? th.lda_dev:rda_dev;
		ldb_dev = (th.ldb_dev!=0) ? th.ldb_dev:rdb_dev;
		ldc_dev = (th.ldc_dev!=0) ? th.ldc_dev:rdc_dev;
		cudaMalloc ((void **) &dev_A, th.word_size * lda_dev * cda_dev);
		cudaMalloc ((void **) &dev_B, th.word_size * ldb_dev * cdb_dev);
		cudaMalloc ((void **) &dev_C, th.word_size * ldc_dev * cdc_dev);
		#endif

		// memcpy from host to device
		cublasSetMatrix(rda_dev, cda_dev, th.word_size, hst_A, lda_hst, dev_A, lda_dev);
		cublasSetMatrix(rdb_dev, cdb_dev, th.word_size, hst_B, ldb_hst, dev_B, ldb_dev);
		cublasSetMatrix(rdc_dev, cdc_dev, th.word_size, hst_C, ldc_hst, dev_C, ldc_dev);

		if (th.mode != 't') {
			printf ("%zd\t%zd\t%zd", (size_t)th.dim_m_dev, (size_t)th.dim_n_dev, (size_t)th.dim_k_dev);
		}
		get_routine_theoretial_performance (&th);

		// execution ---------------------------------
		if (th.mode == 'p' || th.mode == 'e') {
			ROUTINE_TARGET(ha,th,th.dim_m_dev,th.dim_n_dev,th.dim_k_dev,alpha,dev_A,lda_dev,dev_B,ldb_dev,beta,dev_C,ldc_dev); // dummy
			#if PERF_AVERAGE
			float sec;
			iter = get_num_iter (&th);
			t0 = gettime ();
			for (i = 0; i < iter; i++) ROUTINE_TARGET(ha,th,th.dim_m_dev,th.dim_n_dev,th.dim_k_dev,alpha,dev_A,lda_dev,dev_B,ldb_dev,beta,dev_C,ldc_dev);
			t1 = gettime ();
			sec = (t1 - t0) / (float)iter;
			if (th.mode == 'e') printf ("\t%f\t%f", t0, t1);
			printf ("\t%1.3e\t%1.3e\t%1.3e\t%zd", sec, (th.routine_flops / sec) * 1.0e-9, (th.routine_bytes / sec) * 1.0e-9, (size_t)iter);
			#else
			float sec_all_exec[MINLOOP];
			float sec_best;//, sec_worst;
			iter = MINLOOP;
			for (i = 0; i < iter; i++) {
				t0 = gettime ();
				ROUTINE_TARGET(ha,th,th.dim_m_dev,th.dim_n_dev,th.dim_k_dev,alpha,dev_A,lda_dev,dev_B,ldb_dev,beta,dev_C,ldc_dev);
				t1 = gettime ();
				sec_all_exec[i] = (t1 - t0);
			}
			sec_best  = get_best  (sec_all_exec);
			printf ("\t%1.3e\t%1.3e\t%1.3e\t%zd", sec_best, (th.routine_flops / sec_best) * 1.0e-9, (th.routine_bytes / sec_best) * 1.0e-9, (size_t)iter);
//			sec_worst = get_worst (sec_all_exec);
//			printf ("\t%1.3e\t%1.3e\t%1.3e\t%zd", sec_worst, (th.routine_flops / sec_worst) * 1.0e-9, (th.routine_bytes / sec_worst) * 1.0e-9, (size_t)iter);
			#endif
		}
		if (th.mode == 'c' || th.mode == 't') 
			ROUTINE_TARGET(ha,th,th.dim_m_dev,th.dim_n_dev,th.dim_k_dev,alpha,dev_A,lda_dev,dev_B,ldb_dev,beta,dev_C,ldc_dev);
		err = cudaGetLastError();
		if (err != cudaSuccess) printf("error : %s\n", cudaGetErrorString(err));
		// -------------------------------------------

		if (th.mode == 'c') 
			cublasGetMatrix (rdc_dev, cdc_dev, th.word_size, dev_C, ldc_dev, hst_C_t, ldc_hst);

		cudaFree (dev_A);
		cudaFree (dev_B);
		cudaFree (dev_C);

		if (th.mode == 'c') {
			mublasConvert2D (1, 1, 0, &alpha, &alpha_r);
			mublasConvert2D (1, 1, 0, &beta, &beta_r);
			mublasConvert2D (rda_dev, cda_dev, lda_hst, hst_A, hst_A_r);
			mublasConvert2D (rdb_dev, cdb_dev, ldb_hst, hst_B, hst_B_r);
			mublasConvert2D (rdc_dev, cdc_dev, ldc_hst, hst_C, hst_C_r);
			#if (defined (CHECK_CUBLAS))
			REFERENCE_TYPE *dev_A_r, *dev_B_r, *dev_C_r;
			cudaMalloc ((void **) &dev_A_r, word_size_r * lda_dev * cda_dev);
			cudaMalloc ((void **) &dev_B_r, word_size_r * ldb_dev * cdb_dev);
			cudaMalloc ((void **) &dev_C_r, word_size_r * ldc_dev * cdc_dev);
			cublasSetMatrix (rda_dev, cda_dev, word_size_r, hst_A_r, lda_hst, dev_A_r, lda_dev);
			cublasSetMatrix (rdb_dev, cdb_dev, word_size_r, hst_B_r, ldb_hst, dev_B_r, ldb_dev);
			cublasSetMatrix (rdc_dev, cdc_dev, word_size_r, hst_C_r, ldc_hst, dev_C_r, ldc_dev);
			ROUTINE_REFERENCE(ch,th,th.dim_m_dev,th.dim_n_dev,th.dim_k_dev,alpha_r,dev_A_r,lda_dev,dev_B_r,ldb_dev,beta_r,dev_C_r,ldc_dev);
			cublasGetMatrix (rdc_dev, cdc_dev, word_size_r, dev_C_r, ldc_dev, hst_C_r, ldc_hst);
			cudaFree (dev_A_r);
			cudaFree (dev_B_r);
			cudaFree (dev_C_r);
			#else
			ROUTINE_REFERENCE(th,th.dim_m_dev,th.dim_n_dev,th.dim_k_dev,alpha_r,hst_A_r,lda_hst,hst_B_r,ldb_hst,beta_r,hst_C_r,ldc_hst);
			#endif
			mublasCheckVector (th.dim_m_hst * th.dim_n_hst, hst_C_t, hst_C_r);
		}

		if (th.mode != 't') {
			#ifdef MUBLAS
			print_mublas_info1 (&th, &ha);
			#endif
		}
		printf ("\n");

		dim_dev_increment (&th);
	}
// --------------------------------------------

// shutdown -----------------------------------
	cudaFreeHost (hst_A);
	cudaFreeHost (hst_B);
	cudaFreeHost (hst_C);
	if (th.mode == 'c') {
		cudaFreeHost (hst_A_r);
		cudaFreeHost (hst_B_r);
		cudaFreeHost (hst_C_r);
		cudaFreeHost (hst_C_t);
	}

	#if defined (CHECK_CUBLAS)
	cublasDestroy (ch);
	#endif
	#if defined (MUBLAS)
	print_mublas_info2 (&th, &ha);
	mublasDestroy (&ha);
	#elif defined (CUBLAS)
	cublasDestroy (ha);
	#elif defined (KBLAS)
	#elif defined (MAGMA)
	#elif defined (CULA)
	#elif defined (ASPEN)
	ASPEN_shutdown();
	#endif
// --------------------------------------------

#endif
	return 0;
}

