#ifndef TESTING_LEVEL3
#define TESTING_LEVEL3

// === GEMM =====================================================
#if defined (GEMM)
#define DIM3
// target routine
#if defined (MUBLAS)
#define ROUTINE_TARGET(ha,th,m,n,k,al,A,lda,B,ldb,bt,C,ldc) MUBLAS_GEMM(&ha,th.tranA,th.tranB,m,n,k,&al,A,lda,B,ldb,&bt,C,ldc)
#elif defined (CUBLAS)
#if defined (MGPU)
#define ROUTINE_TARGET(ha,th,m,n,k,al,A,lda,B,ldb,bt,C,ldc) CUBLAS_GEMM_MGPU(ha,cublas_trans_const_m(th.tranA),cublas_trans_const_m(th.tranB),m,n,k,&al,A,lda,B,ldb,&bt,C,ldc)
#else
#define ROUTINE_TARGET(ha,th,m,n,k,al,A,lda,B,ldb,bt,C,ldc) CUBLAS_GEMM(ha,cublas_trans_const_m(th.tranA),cublas_trans_const_m(th.tranB),m,n,k,&al,A,lda,B,ldb,&bt,C,ldc)
#endif
#elif defined (KBLAS)
#elif defined (BLASX)
#define ROUTINE_TARGET(ha,th,m,n,k,al,A,lda,B,ldb,bt,C,ldc) BLASX_GEMM_MGPU(ha,cublas_trans_const_m(th.tranA),cublas_trans_const_m(th.tranB),m,n,k,&al,A,lda,B,ldb,&bt,C,ldc)
#elif defined (MAGMA)
#define ROUTINE_TARGET(ha,th,m,n,k,al,A,lda,B,ldb,bt,C,ldc) MAGMA_GEMM(magma_trans_const(th.tranA),magma_trans_const(th.tranB),m,n,k,al,A,lda,B,ldb,bt,C,ldc)
#elif defined (CULA)
#elif defined (ASPEN)
#endif
// reference routine
#if defined (CHECK_MPACK)
#define ROUTINE_REFERENCE(ha,th,m,n,k,al,A,lda,B,ldb,bt,C,ldc) MBLAS_GEMM(th.tranA,th.tranB,m,n,k,&al,A,lda,B,ldb,&bt,C,ldc)
#elif defined (CHECK_CBLAS)
#if defined (PREC_S) || defined (PREC_D) || defined (PREC_W) || defined (PREC_B) || defined (PREC_DB)
#define ROUTINE_REFERENCE(ha,th,m,n,k,al,A,lda,B,ldb,bt,C,ldc) CBLAS_GEMM(CblasColMajor,cblas_trans_const_m(th.tranA),cblas_trans_const_m(th.tranB),m,n,k,al,A,lda,B,ldb,bt,C,ldc)
#else
#define ROUTINE_REFERENCE(ha,th,m,n,k,al,A,lda,B,ldb,bt,C,ldc) CBLAS_GEMM(CblasColMajor,cblas_trans_const_m(th.tranA),cblas_trans_const_m(th.tranB),m,n,k,&al,A,lda,B,ldb,&bt,C,ldc)
#endif
#elif defined (CHECK_CUBLAS)
#define ROUTINE_REFERENCE(ha,th,m,n,k,al,A,lda,B,ldb,bt,C,ldc) CUBLAS_GEMM(ha,cublas_trans_const_m(th.tranA),cublas_trans_const_m(th.tranB),m,n,k,&al,A,lda,B,ldb,&bt,C,ldc)
#endif
// %%%
#define ROUTINE_FLOPS(m,n,k)			(FPMADCOUNT*m*n*k+(2*FPMULCOUNT+FPADDCOUNT)*m*n)
#define ROUTINE_BYTES(m,n,k,s)			((m*k+k*n+2*m*n)*s) 
#define ROUTINE_MAXSIZE(gmem,margin,s)	(sqrt((gmem-margin)/s/3)) 
#define ROUTINE_MAXSIZE_1(gmem,margin,s,dim1)		((gmem-margin)/s/dim1) // !!!!!!!!!!!!!!
#define ROUTINE_MAXSIZE_2(gmem,margin,s,dim1,dim2)	((gmem+margin+s+dim1+dim2)*0)  // !!!!!!!!!!!!!!!
#endif
// ==============================================================

// === GEAM =====================================================
#if defined (GEAM)
#define DIM2
// target routine
#if defined (MUBLAS)
#define ROUTINE_TARGET(ha,th,m,n,k,al,A,lda,B,ldb,bt,C,ldc) MUBLAS_GEAM(&ha,th.tranA,th.tranB,m,n,&al,A,lda,&bt,B,ldb,C,ldc)
#elif defined (CUBLAS)
#define ROUTINE_TARGET(ha,th,m,n,k,al,A,lda,B,ldb,bt,C,ldc) CUBLAS_GEAM(ha,cublas_trans_const_m(th.tranA),cublas_trans_const_m(th.tranB),m,n,&al,A,lda,&bt,B,ldb,C,ldc)
#elif defined (KBLAS)
#elif defined (MAGMA)
#elif defined (CULA)
#elif defined (ASPEN)
#endif
// reference routine
#if defined (CHECK_MPACK)
#elif defined (CHECK_CBLAS)
#elif defined (CHECK_CUBLAS)
#define ROUTINE_REFERENCE(ha,th,m,n,k,al,A,lda,B,ldb,bt,C,ldc) CUBLAS_GEAM(ha,cublas_trans_const_m(th.tranA),cublas_trans_const_m(th.tranB),m,n,&al,A,lda,&bt,B,ldb,C,ldc)
#endif
// %%%
#define ROUTINE_FLOPS(m,n)				((FPADDCOUNT+FPMADCOUNT)*m*n)
#define ROUTINE_BYTES(m,n,s)			((3*m*n)*s) 
#define ROUTINE_MAXSIZE(gmem,margin,s)	(sqrt((gmem-margin)/s/3)) 
#define ROUTINE_MAXSIZE_1(gmem,margin,s,dim1)		((gmem-margin)/s/dim1) // !!!!!!!!!!!!!!
#define ROUTINE_MAXSIZE_2(gmem,margin,s,dim1,dim2)	((gmem+margin+s+dim1+dim2)*0)  // !!!!!!!!!!!!!!!
#endif
// ==============================================================


#endif
